from writer.abstract import register_abstract_template
from writer.blocks.base_block import BlueprintBlock
from writer.ss_types import AbstractTemplate, WriterConfigurationError


class WriterAddChatMessage(BlueprintBlock):

    @classmethod
    def register(cls, type: str):
        super(WriterAddChatMessage, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="blueprints_node",
            writer={
                "name": "Add chat message",
                "description": "Adds a message to the conversation history. Use for displaying user or AI messages.",
                "category": "Writer",
                "deprecated": True,
                "fields": {
                    "conversationStateElement": {
                        "name": "Conversation state element",
                        "desc": "The variable that has your conversation object.",
                        "type": "Binding",
                    },
                    "message": {
                        "name": "Message",
                        "type": "Object",
                        "init": '{ "role": "assistant", "content": "Hello" }',
                        "desc": "A message object. Content can be text string or array of objects for multimodal (text + images) with X5 model.",
                        "validator": {
                            "type": "object",
                            "properties": {
                                "role": {"type": "string"},
                                "content": {
                                    "oneOf": [
                                        {"type": "string"},
                                        {
                                            "type": "array",
                                            "items": {
                                                "oneOf": [
                                                    {
                                                        "type": "object",
                                                        "properties": {
                                                            "type": {"type": "string", "enum": ["text"]},
                                                            "text": {"type": "string"}
                                                        },
                                                        "required": ["type", "text"],
                                                        "additionalProperties": False
                                                    },
                                                    {
                                                        "type": "object",
                                                        "properties": {
                                                            "type": {"type": "string", "enum": ["image_url"]},
                                                            "image_url": {
                                                                "type": "object",
                                                                "properties": {
                                                                    "url": {"type": "string"}
                                                                },
                                                                "required": ["url"],
                                                                "additionalProperties": False
                                                            }
                                                        },
                                                        "required": ["type", "image_url"],
                                                        "additionalProperties": False
                                                    }
                                                ]
                                            }
                                        }
                                    ]
                                },
                            },
                            "additionalProperties": False,
                        }
                    }
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "The message was added to the conversation.",
                        "style": "success",
                    },
                    "error": {
                        "name": "Error",
                        "description": "There was an error adding the message.",
                        "style": "error",
                    },
                },
            }
        ))

    def run(self):
        try:
            import writer.ai

            conversation_state_element = self._get_field("conversationStateElement", required=True)
            message = self._get_field("message", as_json=True, required=True)

            conversation = self.evaluator.evaluate_expression(conversation_state_element, self.instance_path, self.execution_environment)

            if conversation is None or not isinstance(conversation, writer.ai.Conversation):
                raise WriterConfigurationError("The state element specified doesn't contain a conversation. Initialize one using the block 'Start chat conversation'.")
            
            writer.ai.Conversation.validate_message(message)
            conversation += message

            self._set_state(conversation_state_element, conversation)            
            self.result = None
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e

    



from writer.blocks.addtostatelist import AddToStateList
from writer.blocks.apitrigger import APITrigger
from writer.blocks.calleventhandler import CallEventHandler
from writer.blocks.changepage import ChangePage
from writer.blocks.code import CodeBlock
from writer.blocks.crontrigger import CronTrigger
from writer.blocks.foreach import ForEach
from writer.blocks.httprequest import HTTPRequest
from writer.blocks.ifelse import IfElseBlock
from writer.blocks.logmessage import LogMessage
from writer.blocks.parsejson import ParseJSON
from writer.blocks.returnvalue import ReturnValue
from writer.blocks.runblueprint import RunBlueprint
from writer.blocks.setstate import SetState
from writer.blocks.sharedblueprint import SharedBlueprint
from writer.blocks.uieventtrigger import UIEventTrigger
from writer.blocks.writeraddchatmessage import WriterAddChatMessage
from writer.blocks.writeraddtokg import WriterAddToKG
from writer.blocks.writeraskkg import WriterAskGraphQuestion
from writer.blocks.writerchat import WriterChat
from writer.blocks.writerchatreply import WriterChatReply
from writer.blocks.writerchatreplywithtoolconfig import WriterChatReplyWithToolConfig
from writer.blocks.writerclassification import WriterClassification
from writer.blocks.writercompletion import WriterCompletion
from writer.blocks.writerfileapi import WriterUploadFile
from writer.blocks.writerinitchat import WriterInitChat
from writer.blocks.writerkeyvaluestorage import WriterKeyValueStorage
from writer.blocks.writernocodeapp import WriterNoCodeApp
from writer.blocks.writerparsepdf import WriterParsePDFByFileID
from writer.blocks.writerstructuredoutput import WriterStructuredOutput
from writer.blocks.writertoolcalling import WriterToolCalling
from writer.blocks.writervision import WriterVision
from writer.blocks.writerwebsearch import WriterWebSearch

SetState.register("blueprints_setstate")
WriterClassification.register("blueprints_writerclassification")
WriterCompletion.register("blueprints_writercompletion")
WriterChatReply.register("blueprints_writerchatreply")
HTTPRequest.register("blueprints_httprequest")
RunBlueprint.register("blueprints_runblueprint")
WriterNoCodeApp.register("blueprints_writernocodeapp")
ForEach.register("blueprints_foreach")
IfElseBlock.register("blueprints_ifelse")
LogMessage.register("blueprints_logmessage")
WriterChat.register("blueprints_writerchat")
WriterAddChatMessage.register("blueprints_writeraddchatmessage")
ParseJSON.register("blueprints_parsejson")
CallEventHandler.register("blueprints_calleventhandler")
AddToStateList.register("blueprints_addtostatelist")
ReturnValue.register("blueprints_returnvalue")
WriterInitChat.register("blueprints_writerinitchat")
WriterAddToKG.register("blueprints_writeraddtokg")
UIEventTrigger.register("blueprints_uieventtrigger")
APITrigger.register("blueprints_apitrigger")
CronTrigger.register("blueprints_crontrigger")
CodeBlock.register("blueprints_code")
ChangePage.register("blueprints_changepage")
WriterStructuredOutput.register("blueprints_writerstructuredoutput")
WriterToolCalling.register("blueprints_writertoolcalling")
WriterAskGraphQuestion.register("blueprints_writeraskgraphquestion")
WriterParsePDFByFileID.register("blueprints_writerparsepdf")
WriterUploadFile.register("blueprints_writeruploadfile")
WriterVision.register("blueprints_writervision")
WriterWebSearch.register("blueprints_writerwebsearch")
WriterKeyValueStorage.register("blueprints_writerkeyvaluestorage")
WriterChatReplyWithToolConfig.register("blueprints_writerchatreplywithtoolconfig")
SharedBlueprint.register("blueprints_shared")



import inspect

import writer.core
from writer.abstract import register_abstract_template
from writer.blocks.base_block import BlueprintBlock
from writer.ss_types import AbstractTemplate


class CallEventHandler(BlueprintBlock):
    @classmethod
    def register(cls, type: str):
        super(CallEventHandler, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="blueprints_node",
                writer={
                    "name": "Call event handler",
                    "description": "Executes an event handler.",
                    "category": "Logic",
                    "deprecated": True,
                    "fields": {
                        "name": {
                            "name": "Name",
                            "type": "Handler",
                            "desc": "The name of the event handling function.",
                        },
                        "additionalArgs": {
                            "name": "Additional arguments",
                            "init": '{ "my_arg": 2 }',
                            "type": "Object",
                            "control": "Textarea",
                            "default": "{}",
                        },
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "The event handler execution was successful.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "The event handler execution wasn't successful.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            handler_name = self._get_field("name")
            additional_args = self._get_field("additionalArgs", as_json=True)

            current_app_process = writer.core.get_app_process()
            handler_registry = current_app_process.handler_registry
            callable_handler = handler_registry.find_handler_callable(handler_name)

            args = {
                "state": self.runner.session.session_state,
                "context": self.execution_environment,
                "session": writer.core._event_handler_session_info(),
                "ui": writer.core._event_handler_ui_manager(),
            } | additional_args

            handler_args = inspect.getfullargspec(callable_handler).args
            func_args = []
            for arg in handler_args:
                if arg in args:
                    func_args.append(args[arg])
            self.result = callable_handler(*func_args)
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



from typing import TYPE_CHECKING, Any, Dict, Optional, Type

import httpx
from writerai import DefaultHttpxClient, Writer

import writer.core_ui
import writer.evaluator
from writer.ss_types import WriterConfigurationError

if TYPE_CHECKING:
    from writer.blueprints import BlueprintRunner
    from writer.core_ui import Component
    from writer.ss_types import InstancePath

BlueprintBlock_T = Type["BlueprintBlock"]
block_map: Dict[str, BlueprintBlock_T] = {}


class BlueprintBlock:
    _parent_client_class = httpx.Client
    _custom_httpx_client: Optional[httpx.Client] = None
    _log_requests: bool = True

    @classmethod
    def register(cls, type: str):
        block_map[type] = cls

    def __init_subclass__(
        cls,
        custom_httpx_client: Optional['httpx.Client'] = None,
        log_requests: bool = True,
        **kwargs
    ):
        super().__init_subclass__(**kwargs)
        cls._custom_httpx_client = custom_httpx_client
        cls._log_requests = log_requests

    def __init__(
        self,
        component: writer.core_ui.Component,
        runner: "BlueprintRunner",
        execution_environment: Dict,
    ):
        self.outcome: Optional[str] = None
        self.message: Optional[str] = None
        self.component = component
        self.runner = runner
        self.execution_time_in_seconds = -1.0
        self.started_at = -1.0
        self.captured_stdout = ""
        self.captured_logs = ""
        self.execution_environment = execution_environment
        # A stable snapshot of the execution environment taken after the block
        # has finished running. Used when generating logs to avoid concurrent
        # mutations while serialising.
        self.execution_environment_snapshot: Optional[Dict] = None
        self.result = None
        self.return_value = None
        self.instance_path: InstancePath = [{"componentId": component.id, "instanceNumber": 0}]
        self.evaluator = writer.evaluator.Evaluator(
            runner.session.session_state, runner.session.session_component_tree
        )

    def _handle_missing_field(self, field_key):
        field_content = self.component.content.get(field_key)
        if field_content:
            raise WriterConfigurationError(
                f"The field `{field_key}` is required. The expression specified, `{field_content}`, resulted in an empty value."
            )
        else:
            raise WriterConfigurationError(
                f"The field `{field_key}` is required. It was left empty."
            )

    def _get_field(self, field_key: str, as_json=False, default_field_value=None, required=False):
        if default_field_value is None:
            if as_json:
                default_field_value = "{}"
            else:
                default_field_value = ""
        value = self.evaluator.evaluate_field(
            self.instance_path, field_key, as_json, default_field_value, self.execution_environment
        )

        if required and (value is None or value == "" or value == {}):
            self._handle_missing_field(field_key)

        return value

    def _set_state(self, expr: str, value: Any):
        self.evaluator.set_state(
            expr, self.instance_path, value, base_context=self.execution_environment
        )

    def run(self):
        pass

    def create_httpx_client(self, *args, **kwargs) -> httpx.Client:
        """
        Create a custom HTTPX client with request and response logging.
        """
        logger = self.create_logger() if self._log_requests else None

        event_hooks = {
            "request": [logger.request_hook] if logger else [],
            "response": [logger.response_hook] if logger else [],
        }

        def send(self_inner, request, **kw):
            response = self._parent_client_class.send(self_inner, request, **kw)
            log_entry = response.extensions.get("log_entry")
            if log_entry:
                # Store reference to log entry for later update
                response._log_entry_ref = log_entry
                
                try:
                    # For non-streaming responses, capture content immediately
                    if response.is_closed or response.is_stream_consumed:
                        log_entry["response"]["content"] = response.text
                    else:
                        # For streaming responses, we'll update after consumption
                        log_entry["response"]["content"] = "<pending stream consumption>"
                        
                        # Wrap the response read/iter methods to capture content after streaming
                        original_read = response.read
                        original_iter_raw = response.iter_raw
                        original_iter_bytes = response.iter_bytes
                        original_iter_text = response.iter_text
                        original_iter_lines = response.iter_lines
                        
                        def wrapped_read():
                            try:
                                content = original_read()
                            except Exception as e:
                                if hasattr(response, '_log_entry_ref'):
                                    response._log_entry_ref["response"]["content"] = f"<error reading response stream: {e}>"
                                raise
                            else:
                                if hasattr(response, '_log_entry_ref'):
                                    try:
                                        response._log_entry_ref["response"]["content"] = content.decode('utf-8', errors='replace')
                                    except Exception:
                                        response._log_entry_ref["response"]["content"] = "<binary content>"
                                return content
                        
                        def wrapped_iter_raw(*args, **kwargs):
                            chunks = []
                            try:
                                for chunk in original_iter_raw(*args, **kwargs):
                                    chunks.append(chunk)
                                    yield chunk
                            finally:
                                if hasattr(response, '_log_entry_ref'):
                                    try:
                                        response._log_entry_ref["response"]["content"] = b"".join(chunks).decode("utf-8", errors="replace")
                                    except Exception:
                                        response._log_entry_ref["response"]["content"] = "<binary content>"
                        
                        def wrapped_iter_bytes(*args, **kwargs):
                            accumulated_content = []
                            try:
                                for chunk in original_iter_bytes(*args, **kwargs):
                                    accumulated_content.append(chunk)
                                    yield chunk
                            finally:
                                if hasattr(response, '_log_entry_ref'):
                                    try:
                                        full_content = b''.join(accumulated_content)
                                        response._log_entry_ref["response"]["content"] = full_content.decode('utf-8', errors='replace')
                                    except Exception:
                                        response._log_entry_ref["response"]["content"] = "<binary content>"
                        
                        def wrapped_iter_text(*args, **kwargs):
                            text_chunks = []
                            try:
                                for chunk in original_iter_text(*args, **kwargs):
                                    text_chunks.append(chunk)
                                    yield chunk
                            finally:
                                if hasattr(response, '_log_entry_ref'):
                                    response._log_entry_ref["response"]["content"] = ''.join(text_chunks)
                        
                        def wrapped_iter_lines(*args, **kwargs):
                            lines = []
                            try:
                                for line in original_iter_lines(*args, **kwargs):
                                    lines.append(line)
                                    yield line
                            finally:
                                if hasattr(response, '_log_entry_ref'):
                                    response._log_entry_ref["response"]["content"] = '\n'.join(lines)
                        
                        response.read = wrapped_read
                        response.iter_raw = wrapped_iter_raw
                        response.iter_bytes = wrapped_iter_bytes
                        response.iter_text = wrapped_iter_text
                        response.iter_lines = wrapped_iter_lines
                except Exception as e:
                    log_entry["response"]["content"] = f"<error reading response: {e}>"
            return response

        # Create a new class that inherits from the parent client class
        # and overrides the send method to include logging.
        LoggingHttpxClient = type(
            "LoggingHttpxClient",
            (self._parent_client_class,),
            {"send": send}
        )

        return LoggingHttpxClient(
            *args,
            event_hooks=kwargs.pop("event_hooks", {}) | event_hooks,
            **kwargs
        )

    def acquire_httpx_client(self) -> httpx.Client:
        """
        Acquire an HTTPX client for making requests.
        """
        if self._custom_httpx_client:
            return self._custom_httpx_client

        # Create a new HTTPX client with request and response logging
        return self.create_httpx_client()

    def create_logger(
            self,
            env_storage_key: Optional[str] = None
    ):
        import uuid
        instance_path = self.instance_path[0].get('componentId', None)

        class ExecutionEnvironmentLogger:
            """
            This class provides hooks to log requests and responses
            to the execution environment.
            """
            env_storage_key = 'httpx_requests'

            def __init__(
                    self,
                    env: Dict,
                    env_storage_key: Optional[str] = None
            ):
                self.env = env  # Fallback environment

                if env_storage_key:
                    self.env_storage_key = env_storage_key

                # Create a copy of inherited API calls to avoid shared reference issues
                inherited_calls = self.env.get(self.env_storage_key, [])
                self.env[self.env_storage_key] = inherited_calls.copy()
                self.instance_path = instance_path
            
            def _get_current_execution_environment(self) -> Dict:
                """Resolve the currently executing block's environment for logging."""
                try:
                    from writer.blueprints import get_current_block
                    current_block = get_current_block()
                    if current_block and hasattr(current_block, 'execution_environment'):
                        return current_block.execution_environment
                except Exception:
                    pass
                return self.env

            def request_hook(self, request: httpx.Request):
                import datetime as dt
                request_id = str(uuid.uuid4())
                
                # Get the current block's ID for accurate logging
                current_block_id = self.instance_path  # fallback (should be componentId)
                try:
                    from writer.blueprints import get_current_block
                    current_block = get_current_block()
                    if current_block and hasattr(current_block, 'instance_path'):
                        # Extract componentId from the current block's instance path
                        current_block_id = current_block.instance_path[0].get('componentId', None)
                except Exception:
                    pass
                
                # Capture request content
                content = None
                if request.content:
                    try:
                        content = request.content.decode('utf-8', errors='replace')
                    except Exception:
                        content = "<binary content>"
                elif request.stream:
                    # For streaming requests, we can't easily capture the content
                    # without consuming the stream
                    content = "<streaming request body>"
                else:
                    content = None

                log_entry = {
                    'id': request_id,
                    'created_at': dt.datetime.now(dt.timezone.utc).isoformat(),
                    'created_by': current_block_id,
                    'request': {
                        'method': request.method,
                        'url': str(request.url),
                        'headers': dict(request.headers),
                        'content': content
                    },
                    'response': None  # Will populate later
                }

                # Resolve current block's environment for accurate logging
                target_env = self._get_current_execution_environment()
                target_env.setdefault(self.env_storage_key, []).append(log_entry)
                request.extensions['log_entry'] = log_entry

            def response_hook(self, response: httpx.Response):
                log_entry = response.request.extensions.get('log_entry')
                if not log_entry:
                    # Unlikely scenario
                    return

                log_entry['response'] = {
                    'status_code': response.status_code,
                    'url': str(response.url),
                    'headers': dict(response.headers),
                    'content': None  # Initially empty
                }
                response.extensions['log_entry'] = log_entry

        return ExecutionEnvironmentLogger(
            self.execution_environment,
            env_storage_key=env_storage_key
            )


class WriterBlock(BlueprintBlock):
    """
    Base class for all Writer blocks.
    """
    _parent_client_class = DefaultHttpxClient

    def __init__(
        self,
        component: writer.core_ui.Component,
        runner: "BlueprintRunner",
        execution_environment: Dict,
    ):
        super().__init__(component, runner, execution_environment)

        # Initialize the SDK client via block property
        # to set the context and enable logging for AI module downstream.
        # This way, the AI module can use the same HTTPX client
        # without explicitly passing it.
        _ = self.writer_sdk_client

    def _acquire_writer_client(
            self,
            force_new_client: Optional[bool] = False
    ) -> Writer:
        from writer.ai import WriterAIManager

        return WriterAIManager.acquire_client(
            custom_httpx_client=self.acquire_httpx_client(),
            force_new_client=force_new_client
            )

    def create_logger(self, env_storage_key: Optional[str] = "api_calls"):
        return super().create_logger(env_storage_key=env_storage_key)

    @property
    def writer_sdk_client(self) -> Writer:
        return self._acquire_writer_client()



import io
import json
import logging
import logging.config
import os
from contextlib import contextmanager, redirect_stdout
from time import time
from typing import Any, Callable, Dict, List, Optional, Union

WRITER_LOG_LEVEL = os.getenv("WRITER_LOG_LEVEL", "INFO")
WRITER_LOG_FORMAT = os.getenv("WRITER_LOG_FORMAT", "text")  # 'text' or 'json'

FAILOVER_ROUTING_KEY = "unset"
FAILOVER_BUFFER = "failover"


def get_routing_key(prefix: Optional[str] = None) -> str:
    try:
        from writer.blueprints import get_current_block
        current_block = get_current_block()
    except RuntimeError:
        current_block = None

    key = FAILOVER_ROUTING_KEY
    if current_block is not None:
        key = current_block.component.id
    return f"{prefix}-{key}"


class RoutingMap():
    """
    Maintains a map of routing keys to in-memory output buffers (io.StringIO).
    Used for capturing logs or stdout output in different contexts.
    """

    def __init__(self) -> None:
        # It's not expected that this will be used without context.
        # But just in case a fail-over buffer is provided
        self._buffer_map: Dict[str, io.StringIO] = {
            FAILOVER_BUFFER: io.StringIO(),
        }
    
    def get_buffer(self, key: str) -> io.StringIO:
        """
        Retrieve the buffer associated with a routing key.

        If buffer for the key is not present uses a fail-over buffer
        """
        return self._buffer_map.get(key, self._buffer_map[FAILOVER_BUFFER])

    def add_buffer(self, key: str) -> io.StringIO:
        """
        Add a new buffer with the given key.
        """
        buffer = io.StringIO()
        self._buffer_map[key] = buffer
        return buffer

    def remove_buffer(self, key: str) -> None:
        """
        Remove the buffer associated with the specified key.
        """
        self._buffer_map.pop(key, None)


routing_map = RoutingMap()


class RoutingStream(io.StringIO):
    """
    Custom stream that re-routes stdout to the correct io.StringIO buffer
    based on the current context routing key.
    """

    def write(self, s: str) -> int:
        if s.strip():
            logging.getLogger("stdout").info(s)
        routing_key = get_routing_key(prefix="stdout")
        return routing_map.get_buffer(routing_key).write(s)
    
    def getvalue(self) -> str:
        routing_key = get_routing_key(prefix="stdout")
        return routing_map.get_buffer(routing_key).getvalue()


class RoutingHandler(logging.StreamHandler):
    """
    Custom logging handler that re-routes logs to different buffers
    based on the current context routing key.

    Overwritten methods are mirroring original ones from logging.StreamHandler.
    The only difference is how 'stream' object is acquired
    """

    def emit(self, record):
        try:
            msg = self.format(record)
            routing_key = get_routing_key(prefix="logging")
            stream = routing_map.get_buffer(routing_key)
            stream.write(msg + self.terminator)
            self.flush()
        except RecursionError:
            raise
        except Exception:
            self.handleError(record)

    def flush(self):
        routing_key = get_routing_key(prefix="logging")
        stream = routing_map.get_buffer(routing_key)
        with self.lock:
            if stream and hasattr(stream, "flush"):
                stream.flush()


@contextmanager
def use_stdout_redirect(add_log_entry_func: Union[Callable[[str], None], List[Callable[[str], None]]]):
    """
    Context manager that redirects stdout to a context-specific buffer.
    Supports single callback or list of callbacks for multiple output destinations.
    """
    # Normalize to list of callbacks
    callbacks = [add_log_entry_func] if callable(add_log_entry_func) else add_log_entry_func

    key = get_routing_key(prefix="stdout")
    buffer = routing_map.add_buffer(key)
    try:
        with redirect_stdout(RoutingStream()):
            yield buffer
    finally:
        routing_map.remove_buffer(key)
        stdout = buffer.getvalue()
        if stdout:
            for callback in callbacks:
                callback(stdout)


@contextmanager
def use_logging_redirect(add_log_entry_func: Union[Callable[[str], None], List[Callable[[str], None]]]):
    """
    Context manager that redirects logging to a context-specific buffer.
    Supports single callback or list of callbacks for multiple output destinations.
    """
    # Normalize to list of callbacks
    callbacks = [add_log_entry_func] if callable(add_log_entry_func) else add_log_entry_func

    key = get_routing_key(prefix="logging")
    buffer = routing_map.add_buffer(key)
    try:
        yield buffer
    finally:
        routing_map.remove_buffer(key)
        logs = buffer.getvalue()
        if logs:
            for callback in callbacks:
                callback(logs)


class JSONFormatter(logging.Formatter):
    def format(self, record: logging.LogRecord, **kwargs) -> str:
        try:
            from writer.blueprints import get_current_block
            from writer.core import get_app_process, get_session
            current_block = get_current_block()
            app_process = get_app_process()
            session = get_session()
        except RuntimeError:
            current_block = None
            session = None
            app_process = None

        data: Dict[str, Any] = {
            "severity": record.levelname.upper(),
            "message": super().format(record),
            "logger_name": record.name,
            "process": {
                "name": record.processName
            },
            "timestamp": time(),
        }

        if current_block is not None:
            data["component"] = {
                "id": current_block.component.id,
                "type": current_block.component.type
            }

        if session is not None:
            data["session"] = {
                "id": session.session_id,
            }

        if app_process is not None:
            data["process"]["mode"] = app_process.mode

        if isinstance(record.args, dict):
            data.update(record.args)

        # if record.args[0] is a dict add to the json dict
        if isinstance(record.args, tuple):
            if len(record.args) > 0 and isinstance(record.args[0], dict):
                data.update(record.args[0])

        try:
            data_as_json = json.dumps(data)
            return data_as_json 
        except Exception:
            return '{"unserializable": true}' 


def get_handler(format: str = WRITER_LOG_FORMAT, level: str = WRITER_LOG_LEVEL):
    return {
        "level": level,
        "class": "logging.StreamHandler",
        "formatter": format,
    }


def get_logger(level: str = WRITER_LOG_LEVEL):
    return {
        "handlers": ["basic"],
        "level": level,
        "propagate": False,
    }


LOGGING_CONFIG: Dict[str, Any] = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "text": {
            "format": "%(levelname)s - %(name)s - %(message)s",
        },
        "json": {
            "()": JSONFormatter,
        },
        "user": {
            "format": "%(levelname)s - %(message)s"
        }
    },
    "handlers": {
        "basic": get_handler(),
        "stdout": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "json" if WRITER_LOG_FORMAT == "json" else None
        },
        "routing": {
            "()": RoutingHandler,
            "formatter": "user",
        }
    },
    "loggers": {
        "root": get_logger(),
        "writer": get_logger(),
        "app": get_logger(),
        "from_app": get_logger(),
        "journal": get_logger(),
        "kv_storage": get_logger(),
        "vault": get_logger(),
        "exec_logger": {
            "handlers": ["basic", "routing"],
            "level": "DEBUG",
            "propagate": False,
        },
        "user_code": {
            "handlers": ["basic", "routing"],
            "level": "DEBUG",
            "propagate": False,
        },
        "stdout": {
            "handlers": ["stdout"],
            "propagate": False,
        }
    }
}

logging.config.dictConfig(LOGGING_CONFIG)



from json import dumps as json_dumps
from typing import Optional

from writer.core_ui import (
    Component,
    ComponentTree,
    UIError,
    current_component_tree,
    current_parent_container,
)


class WriterUI:
    """Provides mechanisms to manage and manipulate UI components within a
    Writer Framework session.

    This class offers context managers and methods to dynamically create, find,
    and organize UI components based on a structured component tree.
    """
    def __enter__(self):
        return self

    def __exit__(self, *args):
        ...

    @staticmethod
    def assert_in_container():
        container = current_parent_container.get(None)
        if container is None:
            raise UIError("A component can only be created inside a container")

    @property
    def component_tree(self) -> ComponentTree:
        """
        Returns the component tree representation

        :return:
        """
        return current_component_tree()

    @property
    def root(self) -> Component:
        tree = current_component_tree()
        root_component = tree.get_component('root')
        if not root_component:
            raise RuntimeError("Failed to acquire root component")
        return root_component

    @staticmethod
    def find(component_id: str) -> Component:
        """
        Retrieves a component by its ID from the current session's component tree.

        This method searches for a component with the given ID within the
        application's UI structure. If the component is found, it is returned
        for further manipulation or inspection.

        :param component_id: The unique identifier of the component to find.
        :type component_id: str
        :return: The found component with the specified ID.
        :rtype: Component
        :raises RuntimeError: If no component with the specified ID is found
        in the current session's component tree.

        **Example**::

        >>> my_component = ui.find("my-component-id")
        >>> print(my_component.properties)
        """
        # Example context manager for finding components
        component = current_component_tree().get_component(component_id)
        if component is None:
            raise RuntimeError(f"Component {component_id} not found")
        return component

    @staticmethod
    def refresh_with(component_id: str):
        """
        Clears the existing children of a container component and sets it up to
        accept new components. This method is designed to refresh the specified
        container with new content specified in the subsequent block.

        :param component_id: The unique identifier of the container component
                             to be refreshed.
        :raises RuntimeError: If no component with the specified ID is found
        in the current session's component tree.

        .. note:: Upon invocation, this method clears all children of the
        specified container component to prepare for new content. If no new
        components are added within the context block, the container will
        simply be emptied.

        **Example**:
        >>> with ui.refresh_with(id="my-container"):
        >>>     ui.Text({"text": "New content"}, id="new-content-1")
        >>>     ui.Button({"text": "Click me"}, id="new-button-1")

        This method can also be used to clear existing children without adding
        new components:
        >>> with ui.refresh_with(id="my-container"):
        >>>     pass
        """
        component = WriterUI.find(component_id)

        # Clear the children of the specified component.
        current_component_tree().clear_children(component_id)

        return component

    @staticmethod
    def parent(component_id: str, level: int = 1) -> Optional[str]:
        """
        Retrieves the ID of the top-level parent.

        :param component_id:
        :param level:
        :return:
        """
        component_tree = current_component_tree()
        parents_container = component_tree.get_parent(component_id)
        if len(parents_container) < level - 1:
            return None

        return parents_container[level - 1]

    @staticmethod
    def create_container_component(component_type: str, **kwargs) -> Component:
        component_tree = current_component_tree()
        container = _create_component(component_tree, component_type, **kwargs)
        component_tree.attach(container)
        return container

    @staticmethod
    def create_component(component_type: str, **kwargs) -> Component:
        WriterUI.assert_in_container()
        component_tree = current_component_tree()
        component = _create_component(component_tree, component_type, **kwargs)
        component_tree.attach(component)
        return component


def _prepare_handlers(raw_handlers: Optional[dict]):
    handlers = {}
    if raw_handlers is not None:
        for event, handler in raw_handlers.items():
            if callable(handler):
                module_name = \
                    handler.__module__ + "." \
                    if handler.__module__ != "writeruserapp" \
                    else ""
                handlers[event] = \
                    f"{module_name}{handler.__name__}"
            else:
                handlers[event] = handler
    return handlers


def _prepare_binding(raw_binding):
    if raw_binding is not None:
        if len(raw_binding) == 1:
            binding = {
                "eventType": list(raw_binding.keys())[0],
                "stateRef": list(raw_binding.values())[0]
            }
            return binding
        elif len(raw_binding) != 0:
            raise RuntimeError('Improper binding configuration')


def _prepare_value(value):
    if isinstance(value, dict):
        return json_dumps(value)
    return str(value)


def _create_component(component_tree: ComponentTree, component_type: str, **kwargs) -> Component:

    parent_container = current_parent_container.get(None)
    if kwargs.get("id", False) is None:
        kwargs.pop("id")

    if kwargs.get("position", False) is None:
        kwargs.pop("position")

    if kwargs.get("parentId", False) is None:
        kwargs.pop("parentId")

    if "parentId" in kwargs:
        parent_id: str = kwargs.pop("parentId")
    else:
        parent_id = "root" if not parent_container else parent_container.id

    # Converting all passed content values to strings
    raw_content: dict = kwargs.pop("content", {})
    content = {key: _prepare_value(value) for key, value in raw_content.items()}

    # A pre-defined ID is required for page components
    # to prevent page focus loss on app reload
    if component_type == "page" and "id" not in kwargs:
        identifier = f"cmc-page-{component_tree.page_counter + 1}"
        if "key" not in content:
            content["key"] = identifier
        if "id" not in kwargs:
            kwargs["id"] = identifier

    position: Optional[int] = kwargs.pop("position", None)
    is_positionless: bool = kwargs.pop("positionless", False)
    raw_handlers: dict = kwargs.pop("handlers", {})
    raw_binding: dict = kwargs.pop("binding", {})

    handlers = _prepare_handlers(raw_handlers) or None
    binding = _prepare_binding(raw_binding) or None

    component = Component(
        type=component_type,
        parentId=parent_id,
        isCodeManaged=True,
        content=content,
        handlers=handlers,
        binding=binding,
        **kwargs
        )

    # We're determining the position separately
    # due to that we need to know whether parent of the component
    # is present within base component tree
    # or a session-specific one
    component.position = \
        position if position is not None else \
        component_tree.determine_position(
            parent_id,
            is_positionless=is_positionless
            )

    return component



from writer.blocks.base_block import BlueprintBlock
from writer.ss_types import WriterConfigurationError


class BlueprintTrigger(BlueprintBlock):
    def run(self):
        self.result = self.execution_environment.get("payload")

        if not self.result:
            try:
                self.result = self._get_field("defaultResult", True, None)
            except WriterConfigurationError:
                self.result = self._get_field("defaultResult", False, None)




from uuid import uuid4

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WriterBlock
from writer.ss_types import AbstractTemplate


class WriterUploadFile(WriterBlock):

    @classmethod
    def register(cls, type: str):
        super(WriterUploadFile, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="blueprints_node",
            writer={
                "name": "Add files to Writer cloud",
                "description": "Uploads files to the Writer platform.",
                "category": "Writer",
                "fields": {
                    "files": {
                        "name": "Files",
                        "type": "Object",
                        "default": "[]",
                        "desc": "A list of files to be uploaded to the Writer platform. You can use files uploaded via the File input component or specify dictionaries with data, type and name.",
                        "validator": {
                            "type": "array"
                        }
                    },
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "File successfully uploaded.",
                        "style": "success"
                    },
                    "error": {
                        "name": "Error",
                        "description": "There was an error uploading the files.",
                        "style": "error"
                    }
                }
            }
        ))

    def run(self):
        from writer.ai import WriterAIManager
        try:
            files = self._get_field("files", as_json=True, required=True)
            outputs = []
            if not isinstance(files, list):
                raise ValueError("Files must be a list.")
            for file in files:
                if not isinstance(file, dict):
                    raise ValueError("Files must be dictionaries and contain `data`, `type` and `name` attributes.")
                if "data" not in file or "type" not in file:
                    raise ValueError("A file specified as a dictionary must contain `data` and `type` attributes.")

                data = file.get("data")
                file_type = file.get("type")
                name = file.get("name")
                file_name = name or f"agent-builder-{file_type}-{uuid4()}"
                content_disposition = f'attachment; filename="{file_name}"'

                client = WriterAIManager.acquire_client()

                file = client.files.upload(
                    content=data,
                    content_type=file_type,
                    content_disposition=content_disposition
                    )

                outputs.append({
                    "id": file.id,
                    "name": file.name,
                    "status": file.status,
                    "created_at": file.created_at.isoformat(),
                    "graph_ids": file.graph_ids
                })
            if len(outputs) == 0:
                raise ValueError("No files were uploaded.")
            self.result = outputs
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



import json
import logging

from writer.abstract import register_abstract_template
from writer.blocks.base_block import WriterBlock
from writer.ss_types import AbstractTemplate


class WriterAskGraphQuestion(WriterBlock):
    @classmethod
    def register(cls, type: str):
        super(WriterAskGraphQuestion, cls).register(type)
        register_abstract_template(type, AbstractTemplate(
            baseType="blueprints_node",
            writer={
                "name": "Ask graph question",
                "description": "Asks a natural language question using one or more knowledge graphs and puts the result into a state variable.",
                "category": "Writer",
                "fields": {
                    "question": {
                        "name": "Question",
                        "type": "Text",
                        "desc": "The natural language question to ask.",
                        "control": "Textarea",
                    },
                    "useStreaming": {
                        "name": "Use streaming",
                        "type": "Boolean",
                        "default": "yes",
                        "validator": {
                            "type": "boolean",
                        },
                    },
                    "stateElement": {
                        "name": "Link Variable",
                        "type": "Binding",
                        "desc": "Set the variable here and use it across your agent.",
                    },
                    "graphIds": {
                        "name": "Graph Ids",
                        "type": "Graph Ids",
                        "desc": "IDs of the graphs to query.",
                        "default": "",
                        "validator": {
                            "type": "string",
                        }
                    },
                    "subqueries": {
                        "name": "Use subqueries",
                        "type": "Boolean",
                        "desc": "Enables LLM to ask follow-up questions to the knowledge graph. This improves answers, but may be slower.",
                        "default": "yes",
                        "validator": {
                            "type": "boolean",
                        },
                    }
                },
                "outs": {
                    "success": {
                        "name": "Success",
                        "description": "Successfully streamed the answer.",
                        "style": "success"
                    },
                    "error": {
                        "name": "Error",
                        "description": "There was an error answering the question.",
                        "style": "error"
                    }
                }
            }
        ))

    def run(self):
        try:
            client = self.writer_sdk_client

            graph_ids = self._get_field(
                "graphIds", as_json=True, required=True)
            use_streaming = self._get_field(
                "useStreaming", False, "yes") == "yes"
            if isinstance(graph_ids, str):
                graph_ids = [graph_ids]
            elif not isinstance(graph_ids, list):
                raise ValueError(
                    "graphIds must be a string or a list of strings")
            if len(graph_ids) == 0:
                raise ValueError("graphIds must not be empty")

            question = self._get_field("question", required=True)
            state_element = self._get_field("stateElement", required=False)
            if not state_element and use_streaming:
                raise ValueError(
                    "A state element must be provided when using streaming.")
            subqueries = self._get_field(
                "subqueries", default_field_value="yes") == "yes"

            answer_so_far = ""

            response = client.graphs.question(
                graph_ids=graph_ids,
                question=question,
                stream=use_streaming,
                subqueries=subqueries
            )
            if use_streaming:
                for chunk in response:
                    try:
                        delta = chunk.model_extra.get("answer", "")
                        answer_so_far += delta
                        self._set_state(state_element, answer_so_far)
                    except json.JSONDecodeError:
                        logging.error(
                            "Could not parse stream chunk from graph.question")
            else:
                answer_so_far = response.answer
                self._set_state(state_element, answer_so_far)

            self.result = answer_so_far
            self.outcome = "success"

        except BaseException as e:
            self.outcome = "error"
            raise e



"""
Writer Vault module for managing secret retrieval and caching.

This module provides the WriterVault class which handles fetching and caching
secrets from the Writer vault service, with support for environment-based
configuration and error handling.
"""

import logging
from typing import Dict, Optional

import httpx

from writer.keyvalue_storage import writer_kv_storage

logger = logging.getLogger("vault")


class WriterVault:
    """Manages retrieval and caching of secrets from the Writer vault service."""

    def __init__(self) -> None:
        """Initialize vault with empty cache."""
        self.secrets: Optional[Dict] = None

    def get_secrets(self) -> Dict:
        """Get cached secrets, fetching from vault if not already loaded."""
        if self.secrets is None:
            self.secrets = self._fetch()
        return self.secrets

    def refresh(self):
        """Force refresh of secrets from the vault service."""
        self.secrets = self._fetch()

    def _fetch(self) -> Dict:
        try:
            data = writer_kv_storage.get("vault", "secret")
            secrets = data.get("secret")
            if isinstance(secrets, dict):
                return secrets
            logger.warning("Invalid vault response format: expected dict in 'secret' field")
        except (httpx.HTTPStatusError, ValueError) as e:
            logger.error("Failed to fetch vault secrets: %s", e)
        return {}


writer_vault = WriterVault()



from writer.abstract import register_abstract_template
from writer.blocks.base_block import BlueprintBlock
from writer.ss_types import AbstractTemplate


class ChangePage(BlueprintBlock):
    @classmethod
    def register(cls, type: str):
        super(ChangePage, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="blueprints_node",
                writer={
                    "name": "Change page",
                    "description": "Navigates the user to another page in the app. Requires a valid page key.",
                    "category": "Other",
                    "fields": {
                        "pageKey": {
                            "name": "Page key",
                            "type": "Text",
                            "desc": "The identifying key of the target page.",
                        },
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "The page change was successful.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "There was an error changing the page.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            page_key = self._get_field("pageKey", required=True)
            self.runner.session.session_state.set_page(page_key)
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



from writer.abstract import register_abstract_template
from writer.blocks.base_block import WriterBlock
from writer.ss_types import AbstractTemplate

DEFAULT_MODEL = "palmyra-x5"

class WriterCompletion(WriterBlock):
    @classmethod
    def register(cls, type: str):
        super(WriterCompletion, cls).register(type)
        register_abstract_template(
            type,
            AbstractTemplate(
                baseType="blueprints_node",
                writer={
                    "name": "Text generation",
                    "description": "Generates text using a Writer model. Use for completions, summaries, or creative writing.",
                    "category": "Writer",
                    "fields": {
                        "prompt": {"name": "Prompt", "type": "Text", "control": "Textarea"},
                        "modelId": {"name": "Model", "type": "Model Id", "default": DEFAULT_MODEL},
                        "temperature": {
                            "name": "Temperature",
                            "type": "Number",
                            "default": "0.7",
                            "validator": {
                                "type": "number",
                                "minimum": 0,
                                "maximum": 1,
                            },
                        },
                        "max_tokens": {
                            "name": "Max output tokens",
                            "type": "Number",
                            "default": "1024",
                            "validator": {
                                "type": "number",
                                "minimum": 1,
                                "maximum": 16384,
                            }
                        }
                    },
                    "outs": {
                        "success": {
                            "name": "Success",
                            "description": "Text was generated successfully.",
                            "style": "success",
                        },
                        "error": {
                            "name": "Error",
                            "description": "There was an error generating text.",
                            "style": "error",
                        },
                    },
                },
            ),
        )

    def run(self):
        try:
            import writer.ai

            prompt = self._get_field("prompt")
            temperature = float(self._get_field("temperature", False, "0.7"))
            model_id = self._get_field("modelId", False, default_field_value=DEFAULT_MODEL)
            max_tokens = int(self._get_field("max_tokens", False, "1024"))
            config = {"temperature": temperature, "model": model_id, "max_tokens": max_tokens}
            result = writer.ai.complete(prompt, config).strip()
            self.result = result
            self.outcome = "success"
        except BaseException as e:
            self.outcome = "error"
            raise e



import json
import os
import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

import writer.core
import writer.core_ui
from writer.ss_types import (
    InstancePath,
    WriterConfigurationError,
)

if TYPE_CHECKING:
    from writer.core import WriterState
    from writer.core_ui import ComponentTree


class Evaluator:
    """
    Evaluates templates and expressions in the backend.
    It allows for the sanitisation of frontend inputs.
    """

    TEMPLATE_REGEX = re.compile(r"[\\]?@{([^{]*?)}")
    CONTROL_CHARS = re.compile(r"[\x00-\x1f\x7f]")

    def __init__(self, state: "WriterState", component_tree: "ComponentTree"):
        self.state = state
        self.component_tree = component_tree
        self.serializer = writer.core.StateSerialiser()

    def evaluate_field(
        self,
        instance_path: InstancePath,
        field_key: str,
        as_json=False,
        default_field_value="",
        base_context={},
    ) -> Any:
        def decode_json(text):
            if not isinstance(text, str):
                return text
            try:
                # Remove control chars
                clean_text = Evaluator.CONTROL_CHARS.sub("", text)
                return json.loads(clean_text, strict=False)
            except json.JSONDecodeError as exception:
                raise WriterConfigurationError(
                    "Error decoding JSON. " + str(exception)
                ) from exception

        component_id = instance_path[-1]["componentId"]
        component = self.component_tree.get_component(component_id)
        if not component:
            raise ValueError(f'Component with id "{component_id}" not found.')

        field_value = component.content.get(field_key) or default_field_value
        full_match = self.TEMPLATE_REGEX.fullmatch(field_value)

        def replacer(matched: re.Match):
            if matched.group(0)[0] == "\\":  # Escaped @, don't evaluate
                return matched.group(0)
            expr = matched.group(1).strip()
            expr_value = self.evaluate_expression(expr, instance_path, base_context)
            if full_match is not None:
                return expr_value
            if as_json:
                dumped = expr_value
                if not isinstance(dumped, str):
                    dumped = json.dumps(dumped)
                else:
                    dumped = json.dumps(dumped)[1:-1]
                return re.sub(r'(?<!\\)"', r'\"', dumped)
            if not isinstance(expr_value, str):
                return json.dumps(expr_value)
            return expr_value

        if full_match is None:
            replaced = self.TEMPLATE_REGEX.sub(replacer, field_value)
        else:
            replaced = replacer(full_match)
        if as_json:
            replaced = decode_json(replaced)

        return replaced

    def get_context_data(self, instance_path: InstancePath, base_context={}) -> Dict[str, Any]:
        context: Dict[str, Any] = base_context
        for i in range(len(instance_path)):
            path_item = instance_path[i]
            component_id = path_item["componentId"]
            component = self.component_tree.get_component(component_id)
            if not component:
                continue
            if component.type != "repeater":
                continue
            if i + 1 >= len(instance_path):
                continue
            repeater_instance_path = instance_path[0 : i + 1]
            next_instance_path = instance_path[0 : i + 2]
            instance_number = next_instance_path[-1]["instanceNumber"]
            repeater_object = self.evaluate_field(
                repeater_instance_path,
                "repeaterObject",
                True,
                """{ "a": { "desc": "Option A" }, "b": { "desc": "Option B" } }""",
            )
            key_variable = self.evaluate_field(
                repeater_instance_path, "keyVariable", False, "itemId"
            )
            value_variable = self.evaluate_field(
                repeater_instance_path, "valueVariable", False, "item"
            )

            repeater_items: List[Tuple[Any, Any]] = []
            if isinstance(repeater_object, dict):
                repeater_items = list(repeater_object.items())
            elif isinstance(repeater_object, list):
                repeater_items = list(enumerate(repeater_object))
            else:
                raise ValueError(
                    "Cannot produce context. Repeater object must evaluate to a dictionary."
                )

            context[key_variable] = repeater_items[instance_number][0]
            context[value_variable] = repeater_items[instance_number][1]

        if len(instance_path) > 0:
            context["target"] = instance_path[-1]["componentId"]

        return context

    def set_state(
        self, expr: str, instance_path: InstancePath, value: Any, base_context={}
    ) -> None:
        accessors = self.parse_expression(expr, instance_path, base_context)
        state_ref = self.state

        for accessor in accessors[:-1]:
            if isinstance(state_ref, list):
                state_ref = state_ref[int(accessor)]
            else:
                state_ref = state_ref[accessor]

        if not isinstance(
            state_ref, (writer.core.State, writer.core.WriterState, writer.core.StateProxy, dict)
        ):
            raise ValueError(
                f'Reference "{expr}" cannot be translated to state. Found value of type "{type(state_ref)}".'
            )

        state_ref[accessors[-1]] = value

    def parse_expression(
        self, expr: str, instance_path: Optional[InstancePath] = None, base_context={}
    ) -> List[str]:
        """Returns a list of accessors from an expression."""

        if not isinstance(expr, str):
            raise ValueError(
                f'Expression must be of type string. Value of type "{ type(expr) }" found.'
            )

        accessors: List[str] = []
        s = ""
        level = 0

        i = 0
        while i < len(expr):
            character = expr[i]
            if character == "\\":
                if i + 1 < len(expr):
                    s += expr[i + 1]
                    i += 1
            elif character == ".":
                if level == 0:
                    accessors.append(s)
                    s = ""
                else:
                    s += character
            elif character == "[":
                if level == 0:
                    accessors.append(s)
                    s = ""
                else:
                    s += character
                level += 1
            elif character == "]":
                level -= 1
                if level == 0:
                    s = str(self.evaluate_expression(s, instance_path, base_context))
                else:
                    s += character
            else:
                s += character

            i += 1

        if s:
            accessors.append(s)

        return accessors

    def get_env_variable_value(self, expr: str):
        return os.getenv(expr[1:])

    def evaluate_expression(
        self, expr: str, instance_path: Optional[InstancePath] = None, base_context={}
    ) -> Any:
        context_data = base_context
        result = None
        if instance_path:
            context_data = self.get_context_data(instance_path, base_context)
        context_ref: Any = context_data
        state_ref: Any = self.state.user_state
        accessors: List[str] = self.parse_expression(expr, instance_path, base_context)

        result = self._apply_accessors(accessors, state_ref, context_ref)

        if isinstance(result, writer.core.StateProxy):
            return result.to_dict()

        if result is None and expr.startswith("$"):
            return self.get_env_variable_value(expr)

        return result

    def _apply_accessors(self, accessors: List[str], state_ref: Any, context_ref: Any = None) -> Any:
        if not accessors:
            return state_ref
        
        result = self._apply_accessor(accessors[0], context_ref)
        if result is None:
            result = self._apply_accessor(accessors[0], state_ref)

        for accessor in accessors[1:]:
            result = self._apply_accessor(accessor, result)

        return result

    def _apply_accessor(self, accessor: str, target: Any) -> Any:
        if isinstance(target, (writer.core.StateProxy, dict)):
            return target.get(accessor)
        
        if isinstance(target, list):
            try:
                return target[int(accessor)]
            except IndexError:
                pass
        
        return None


import hashlib
import itertools
import json
import logging
import os
import threading
import time
from concurrent.futures import FIRST_COMPLETED, Future, ThreadPoolExecutor, wait
from contextlib import contextmanager
from contextvars import ContextVar, copy_context
from typing import Any, Dict, Generator, List, Literal, Optional, Union

import writer.blocks
import writer.blocks.base_block
import writer.core
import writer.core_ui
from writer.journal import use_journal_record_context
from writer.ss_types import BlueprintExecutionError, BlueprintExecutionLog, WriterConfigurationError

MAX_DAG_DEPTH = 32
MAX_LOG_ITERABLE_SIZE = 100
MAX_LOG_STRING_LENGTH = 5000

_current_block: ContextVar[Optional[writer.blocks.base_block.BlueprintBlock]] = \
    ContextVar("current_block", default=None)


class BlueprintRunManager:
    def __init__(self):
        self._runs: Dict[str, Dict] = {}
        self._lock = threading.Lock()

    @contextmanager
    def register(self, run_id: str):
        event = self.register_run(run_id)
        try:
            yield event
        finally:
            self.deregister_run(run_id)

    def register_run(self, run_id: str):
        with self._lock:
            if run_id not in self._runs:
                self._runs[run_id] = {"counter": 0, "event": threading.Event()}
            self._runs[run_id]["counter"] += 1
            event = self._runs[run_id]["event"]
        return event

    def deregister_run(self, run_id: str):
        with self._lock:
            if run_id in self._runs:
                self._runs[run_id]["counter"] -= 1
                if self._runs[run_id]["counter"] <= 0:
                    del self._runs[run_id]

    def cancel_run(self, run_id: str):
        with self._lock:
            if run_id in self._runs:
                self._runs[run_id]["event"].set()


class BlueprintRunner:
    def __init__(self, session: writer.core.WriterSession):
        self.session = session
        self.executor_lock = threading.Lock()
        self.run_manager = BlueprintRunManager()

    @property
    def api_blueprints(self):
        return self._gather_api_blueprints()

    @property
    def cron_blueprints(self):
        return self._gather_cron_blueprints()

    @contextmanager
    def _get_executor(self) -> Generator[ThreadPoolExecutor, None, None]:
        """Return the application's thread pool executor.

        In normal operation we reuse the main executor provided by the running
        application process. In situations where that process is unavailable
        (for example during tests) a temporary executor is created.
        """

        new_executor = None
        try:
            try:
                current_app_process = writer.core.get_app_process()
                executor = current_app_process.executor
            except RuntimeError:
                logging.info(
                    "The main pool executor isn't being reused. This is only expected in test or debugging situations."
                )
                new_executor = ThreadPoolExecutor(20)  # New executor for debugging/testing
                executor = new_executor

            if not executor:
                raise RuntimeError(
                    "The main pool executor isn't available. This is only expected in test or debugging situations."
                )
            yield executor
        finally:
            if new_executor:
                new_executor.shutdown()

    def execute_ui_trigger(
        self, ref_component_id: str, ref_event_type: str, execution_environment: Dict = {}
    ):
        components = self.session.session_component_tree.get_descendents("blueprints_root")
        ui_triggers = list(filter(lambda c: c.type == "blueprints_uieventtrigger", components))
        for trigger in ui_triggers:
            if trigger.content.get("refComponentId") != ref_component_id:
                continue
            if trigger.content.get("refEventType") != ref_event_type:
                continue
            self.run_branch(trigger.id, None, execution_environment, "UI trigger execution")

    def run_blueprint_by_key(self, blueprint_key: str, execution_environment: Dict = {}):
        all_components = self.session.session_component_tree.components.values()
        blueprints = list(
            filter(
                lambda c: c.type == "blueprints_blueprint" and c.content.get("key") == blueprint_key,
                all_components,
            )
        )
        if len(blueprints) == 0:
            raise ValueError(f'Blueprint with key "{blueprint_key}" not found.')
        blueprint = blueprints[0]
        return self.run_blueprint(
            blueprint.id, execution_environment, f"Blueprint execution ({blueprint_key})"
        )

    def is_blueprint_api_available(
        self, blueprint_id: str
    ):
        """
        Checks if a blueprint with the given key is available for API execution.

        :param blueprint_id: The blueprint identifier.
        :return: True if the blueprint is available for API execution, False otherwise.
        """
        return blueprint_id in self.api_blueprints

    def is_blueprint_cron_available(
        self, blueprint_id: str
    ):
        """
        Checks if a blueprint with the given key is available for Cron execution.

        :param blueprint_id: The blueprint identifier.
        :return: True if the blueprint is available for Cron execution, False otherwise.
        """
        return blueprint_id in self.cron_blueprints

    def get_blueprint_api_trigger(
        self, blueprint_id: str
    ):
        """
        Retrieves the API trigger for a given blueprint key.

        :param blueprint_key: The blueprint identifier.
        :return: The API trigger component.
        """
        if not self.is_blueprint_api_available(blueprint_id):
            raise ValueError(
                f'API trigger not found for blueprint "{blueprint_id}".'
            )
        return self.api_blueprints[blueprint_id]

    def get_blueprint_cron_trigger(
        self, blueprint_id: str
    ):
        """
        Retrieves the Cron trigger for a given blueprint key.

        :param blueprint_key: The blueprint identifier.
        :return: The Cron trigger component.
        """
        if not self.is_blueprint_cron_available(blueprint_id):
            raise ValueError(
                f'Cron trigger not found for blueprint "{blueprint_id}".'
            )
        return self.cron_blueprints[blueprint_id]

    def _gather_blueprints_by_trigger(self, trigger_type: str):
        """
        Gathers all blueprints that have a trigger of the specified type.

        :param trigger_type: The trigger component type (e.g., "blueprints_apitrigger").
        :return: A dict mapping blueprint IDs to their trigger IDs.
        """
        triggers = [
            c for c in self.session.session_component_tree.components.values()
            if c.type == trigger_type
        ]
        blueprints = {}

        for trigger in triggers:
            parent_blueprint_id = \
                self.session.session_component_tree.get_parent(trigger.id)[0]
            parent_blueprint = \
                self.session.session_component_tree.get_component(
                    parent_blueprint_id
                )

            if (
                parent_blueprint
                and
                parent_blueprint.type == "blueprints_blueprint"
            ):
                # Store the blueprint key against its trigger ID
                blueprints[parent_blueprint_id] = trigger.id

        return blueprints

    def _gather_api_blueprints(self):
        """
        Gathers all blueprints that have an API trigger.

        :return: A dict mapping blueprint IDs to their API trigger IDs.
        """
        return self._gather_blueprints_by_trigger("blueprints_apitrigger")

    def _gather_cron_blueprints(self):
        """
        Gathers all blueprints that have a Cron trigger.

        :return: A dict mapping blueprint IDs to their Cron trigger IDs.
        """
        return self._gather_blueprints_by_trigger("blueprints_crontrigger")

    def run_blueprint_via_api(
        self,
        blueprint_id: str,
        branch_id: Optional[str] = None,
        execution_environment: Optional[Dict[str, Any]] = None
    ):
        """
        Executes a blueprint by its key via the API.

        :param blueprint_id: The blueprint identifier.
        :param branch_id: Optional branch ID to start execution from.
        :param execution_environment: The execution environment for
        the blueprint.
        :return: The result of the blueprint execution.
        """
        if execution_environment is None:
            execution_environment = {}

        trigger_id = branch_id
        if trigger_id is not None:
            # Determine trigger type from the component
            component = self.session.session_component_tree.get_component(trigger_id)
            if component and component.type == "blueprints_apitrigger":
                trigger_type = "API"
            elif component and component.type == "blueprints_crontrigger":
                trigger_type = "Cron"
            else:
                trigger_type = "Branch"
        elif self.is_blueprint_api_available(blueprint_id):
            # Prioritize API trigger over Cron if both exist
            trigger_id = self.get_blueprint_api_trigger(blueprint_id)
            trigger_type = "API"
        elif self.is_blueprint_cron_available(blueprint_id):
            trigger_id = self.get_blueprint_cron_trigger(blueprint_id)
            trigger_type = "Cron"
        else:
            raise ValueError(f'No trigger found for blueprint "{blueprint_id}".')
        return self.run_branch(
            trigger_id,
            None,
            execution_environment,
            f"{trigger_type} trigger execution ({blueprint_id} -> {trigger_id})"
        )

    def run_blueprint_batch(self, blueprint_key: str, execution_environments: List[Dict]):
        """
        Executes the same blueprint multiple times sequentially with different execution environments.

        :param blueprint_key: The blueprint identifier (same blueprint for all executions).
        :param execution_environments: A list of execution environments, one per execution.
        :return: A list of results in the same order as execution_environments.
        """
        results = []
        for env in execution_environments:
            result = self.run_blueprint_by_key(blueprint_key, env)
            results.append(result)

        return results

    def _get_blueprint_nodes(self, component_id):
        current_node_id = component_id
        while current_node_id is not None:
            node = self.session.session_component_tree.get_component(current_node_id)
            if not node:
                break
            if node.type == "blueprints_blueprint":
                nodes = self.session.session_component_tree.get_descendents(current_node_id)
                return [node for node in nodes if node.type != 'note']
            current_node_id = node.parentId
        return []

    def run_branch(
        self,
        start_node_id: str,
        branch_out_id: Optional[str],
        execution_environment: Dict,
        title: str = "Branch execution",
    ):
        builder = GraphBuilder(
            components=self._get_blueprint_nodes(start_node_id),
            tools=writer.blocks.base_block.block_map
        )
        if branch_out_id is None:
            builder.set_start_node(start_node_id)
        else:
            builder.set_start_edge(start_node_id, branch_out_id)

        return GraphRunner(
            builder.build(),
            execution_environment, self, title=title
        ).run()

    def run_branch_batch(
        self, base_component_id: str, base_outcome: str, execution_environments: List[Dict]
    ):
        """
        Executes the same branch multiple times sequentially with different execution environments.
        """
        results = []
        for env in execution_environments:
            result = self.run_branch(base_component_id, base_outcome, env)
            results.append(result)

        return results

    def run_blueprint(
        self, component_id: str, execution_environment: Dict, title="Blueprint execution"
    ):
        builder = GraphBuilder(
            components=self._get_blueprint_nodes(component_id),
            tools=writer.blocks.base_block.block_map
        )

        return GraphRunner(
            builder.build(),
            execution_environment, self, title=title
        ).run()

    def cancel_blueprint_execution(self, run_id: str):
        self.run_manager.cancel_run(run_id)


class GraphNode:
    tool_class: writer.blocks.base_block.BlueprintBlock_T
    component: writer.core_ui.Component
    future: Optional[Future] = None
    tool: Optional[writer.blocks.base_block.BlueprintBlock] = None
    # filrered lists of inputs and outputs with only edges from graph
    inputs: List[Any]
    outputs: List[Any]
    status: Optional[str] = None
    _message: Optional[str] = None

    def __init__(self, component: writer.core_ui.Component, graph: "Graph"):
        self.component = component
        self.graph = graph
        tool_class = graph.tools.get(component.type)
        self.inputs = []
        self.outputs = []
        if not tool_class:
            raise WriterConfigurationError(
                f"Component type '{component.type}' is not registered as a block."
            )
        self.tool_class = tool_class


    @property
    def id(self) -> str:
        return self.component.id

    @property
    def result(self) -> Optional[Union[str, Dict]]:
        if self.tool:
            return self.tool.result
        return None

    @property
    def outcome(self) -> Optional[str]:
        if self.status:
            return self.status
        if self.tool:
            return self.tool.outcome
        return None

    @property
    def message(self) -> Optional[str]:
        if self._message:
            return self._message
        if self.tool:
            return self.tool.message
        return None

    @message.setter
    def message(self, value: str):
        self._message = value

    @property
    def return_value(self) -> Optional[Any]:
        if self.tool:
            return self.tool.return_value
        return None

    def run_tool(self, tool: writer.blocks.base_block.BlueprintBlock) -> "GraphNode":
        start_time = time.time()
        tool.started_at = start_time

        call_stack = tool.execution_environment.get("call_stack", []) + [self.id]
        call_depth = call_stack.count(tool.component.id)
        if call_depth > MAX_DAG_DEPTH:
            error_message = f"Maximum call depth ({MAX_DAG_DEPTH}) exceeded. Check that you don't have any unintended circular references."
            tool.outcome = "error"
            tool.message = error_message
            raise RuntimeError(error_message)
        tool.execution_environment["call_stack"] = call_stack
        tool.execution_environment["trace"] = []

        try:
            tool.outcome = "in_progress"
            with use_current_block(tool):
                tool.run()
            if self.outcome == "stopped":
                return self
            tool.outcome = tool.outcome or "success"
        except BlueprintExecutionError as e:
                raise e
        except BaseException as e:
            if not tool.outcome or tool.outcome == "in_progress":
                tool.outcome = "error"
            if isinstance(e, WriterConfigurationError):
                tool.message = str(e)
            else:
                tool.message = repr(e)
            if self._is_error_handled(tool.component, tool.outcome):
                return self 
            else:
                raise e
        finally:
            tool.execution_time_in_seconds = time.time() - start_time
            try:
                tool.execution_environment_snapshot = {
                    k: v for k, v in tool.execution_environment.items() if k != "vault"
                }
            except Exception:
                # pragma: no cover - best effort defensive code
                logging.debug(
                    "Couldn't snapshot execution environment", exc_info=True
                )

        return self 

    def _is_error_handled(self, component: writer.core_ui.Component, outcome: str) -> bool:
        if not component.outs:
            return False
        for output in component.outs:
            if output.get("outId") == outcome:
                return True
        return False

    def _get_env(self, execution_environment: Dict) -> Dict:
        env = execution_environment.copy()
        for inputs in self.inputs:
            from_node = self.graph.get_node(inputs["fromNodeId"])
            if from_node and from_node.tool:
                out_id = inputs.get("outId")
                if out_id and from_node.outcome == out_id:
                    result = from_node.result
                    env['call_stack'] = from_node.tool.execution_environment.get('call_stack', [])
                    env['result'] = result
                    env['message'] = from_node.tool.message
                    
                    # Pass through accumulated API calls from previous block (like call_stack)
                    env['api_calls'] = from_node.tool.execution_environment.get('api_calls', [])
                    env['httpx_requests'] = from_node.tool.execution_environment.get('httpx_requests', [])
        env['results'] = self.graph.get_results()
        return env

    def can_run(self) -> bool:
        if not self.inputs:
            return True
        # all inputs must be evaluated
        for input in self.inputs:
            from_node = self.graph.get_node(input["fromNodeId"])
            if not from_node or from_node.outcome is None or from_node.outcome == "in_progress":
                return False
        return True

    def _is_skipped(self) -> bool:
        if not self.inputs:
            return False
        for input in self.inputs:
            from_node = self.graph.get_node(input["fromNodeId"])
            if from_node and from_node.outcome == input.get("outId"):
                return False
        return True

    def run(self, execution_environment: Dict, runner, executor) -> Future:
        if self.outcome is not None or self._is_skipped():
            self.status = "skipped"
            future: Future = Future()
            future.set_result(self)
            return future

        self.tool = self.tool_class(self.component, runner, self._get_env(execution_environment))
        self.tool.outcome = "in_progress"
        ctx = copy_context()
        self.future = executor.submit(ctx.run, self.run_tool, self.tool)
        if not isinstance(self.future, Future):
            raise WriterConfigurationError(
                f"Unable to run tool {self.tool.component.id} - the executor did not return a Future."
            )
        return self.future

    def debug_info(self) -> Dict[str, Any]:
        return {
            "id": self.id,
            "type": self.component.type,
            "outputs": self.outputs,
            "inputs": self.inputs,
            "result": self.result,
            "message": self.message if self.tool else None,
            "return_value": self.tool.return_value if self.tool else None,
            "outcome": self.outcome if self.tool else None,
        }

class Graph:
    status: Optional[str] = None
    def __init__(self, 
        nodes: List[writer.core_ui.Component],
        tools: Dict[str, writer.blocks.base_block.BlueprintBlock_T]
    ):
        self.tools = tools
        self.nodes = [GraphNode(node, self) for node in nodes]
        self.node_map = {node.id: node for node in self.nodes}
        self._calculate_io()
        self.start_nodes = self._find_start_nodes()

    def get_start_nodes(self) -> List[GraphNode]:
        return self.start_nodes

    def get_node(self, node_id: str) -> Optional[GraphNode]:
        return self.node_map.get(node_id)

    def get_results(self) -> Dict[str, Any]:
        results = {}
        for node in self.nodes:
            if node.tool and node.tool.outcome in ["success", "trigger"]:
                results[node.id] = node.result
        return results

    def _calculate_io(self):
        for node in self.nodes:
            if not node.component.outs:
                continue
            for output in node.component.outs:
                if output.get("toNodeId") in self.node_map:
                    node.outputs.append(output)
                    target_node = self.node_map[output.get("toNodeId")]
                    target_node.inputs.append({
                        "fromNodeId": node.id,
                        "outId": output.get("outId"),
                    })

    def _find_start_nodes(self) -> List[GraphNode]:
        start_nodes = []
        output_nodes = set()
        for node in self.nodes:
            if node.outputs:
                for output in node.outputs:
                    output_nodes.add(output.get("toNodeId"))

        for node in self.nodes:
            if node.id not in output_nodes:
                start_nodes.append(node)

        return start_nodes
    def debug_info(self) -> Dict[str, Any]:
        return {
            "nodes": [node.debug_info() for node in self.nodes],
            "start_nodes": [node.id for node in self.start_nodes],
            "tools": list(self.tools.keys()),
        }

class GraphBuilder:
    def __init__(self, components: List[writer.core_ui.Component], tools: Dict[str, writer.blocks.base_block.BlueprintBlock_T]):
        self.components = components 
        self.tools = tools
        self.start_ids: List[str] = []

    def set_start_node(self, start_component_id: str):
        self.start_ids.append(start_component_id)

    def set_start_edge(self, component_id: str, out_id: str):
        for component in self.components:
            if component.id == component_id:
                if not component.outs:
                    break
                for out in component.outs:
                    if out.get("outId") == out_id:
                        self.start_ids.append(out.get("toNodeId"))

    def validate_graph(self, graph: Graph):
        """Validates the graph for cycles and unreachable nodes and marks them as errors."""
        visited = set()
        stack = set()
        has_cycle = False

        def visit(node: GraphNode):
            if node.status == "error":
                return
            if node.id in stack:
                node.status = "error"
                node.message = "Circular dependency detected."
                nonlocal has_cycle
                has_cycle = True
            if node.id in visited:
                return
            visited.add(node.id)
            stack.add(node.id)
            for output in node.outputs:
                next_node = graph.get_node(output["toNodeId"])
                if next_node:
                    visit(next_node)
            stack.remove(node.id)

        for node in graph.nodes:
            if node.id not in visited:
                visit(node)

        for node in graph.nodes:
            if node.id not in visited:
                node.status = "error"
        if has_cycle:
            graph.status = "error"

    def build(self) -> Graph:
        graph = Graph(self._filter_components(), self.tools)
        self.validate_graph(graph)
        return graph

    def _filter_components(self) -> List[writer.core_ui.Component]:
        if not self.start_ids:
            return self.components
        component_map = {component.id: component for component in self.components}
        # todo: remove duplicates
        filtered_components = set()
        queue = [component_map[component_id] for component_id in self.start_ids if component_id in component_map] 
        while queue:
            component = queue.pop(0)
            filtered_components.add(component.id)
            if component.outs is None:
                continue
            for out in component.outs:
                next_component_id = out["toNodeId"]
                if next_component_id in component_map and next_component_id not in filtered_components:
                    queue.append(component_map[next_component_id])

        return [ 
            component_map[component_id] for component_id in filtered_components
            if component_id in component_map
        ]


class StatusLogger:
    def __init__(self,
        graph: Graph,
        runner: BlueprintRunner,
        run_id: str,
        title: str = "Blueprint execution"
    ):
        self.runner = runner
        self.graph = graph
        self.title = title
        self.run_id = run_id
        self.log_id = self._generate_run_id()
        self.lock = threading.Lock()

    def log(
        self,
        msg: str = "",
        entry_type: Literal["info", "error"] = "info",
        exit: Optional[str] = None,
    ):
        if not writer.core.Config.is_mail_enabled_for_log:
            return
        log_id = self.log_id
        exec_log: BlueprintExecutionLog = BlueprintExecutionLog(runId=self.run_id, summary=[], exit=exit)
        for node in self.graph.nodes:
            #print(node.debug_info())
            if node.tool is None:
                if node.outcome is None:
                    exec_log.summary.append({"componentId": node.id })
                    continue
                else:
                    exec_log.summary.append({
                        "componentId": node.id,
                        "outcome": node.outcome,
                        "message": node.message,
                        "result": None,
                        "returnValue": None,
                        "executionEnvironment": {},
                        "executionTimeInSeconds": 0,
                    })
                    continue
            if node.outcome == "stopped":
                exec_log.summary.append(
                    {
                        "componentId": node.id,
                        "outcome": node.outcome,
                        "message": node.message,
                        "executionTimeInSeconds": node.tool.execution_time_in_seconds,
                    }
                )
                continue
            if node.outcome == "in_progress":
                exec_log.summary.append(
                    {
                        "componentId": node.id,
                        "outcome": node.outcome,
                        "message": node.message,
                        "executionTimeInSeconds": node.tool.execution_time_in_seconds,
                    }
                )
                continue

            exec_log.summary.append(
                {
                    "componentId": node.id,
                    "outcome": node.outcome,
                    "message": node.message,
                    "result": self._summarize_data_for_log(node.result),
                    "returnValue": self._summarize_data_for_log(node.return_value),
                    "executionEnvironment": self._summarize_data_for_log(getattr(node.tool, "execution_environment_snapshot", None)),
                    "executionTimeInSeconds": node.tool.execution_time_in_seconds,
                }
            )
        self.runner.session.session_state.add_log_entry(
            entry_type, self.title, msg, blueprint_execution=exec_log, id=log_id
        )

    def _generate_run_id(self):
        timestamp = str(int(time.time() * 1000))
        salt = os.urandom(8).hex()
        raw_id = f"{self.runner.session.session_id}_{timestamp}_{salt}"
        hashed_id = hashlib.sha256(raw_id.encode()).hexdigest()[:24]
        return hashed_id

    def _summarize_data_for_log(self, data):
        """Convert arbitrary data into a log friendly representation."""

        if data is None:
            return None

        if isinstance(data, list):
            return [self._summarize_data_for_log(item) for item in data[:MAX_LOG_ITERABLE_SIZE]]
        if isinstance(data, dict):
            return {
                k: self._summarize_data_for_log(v)
                for k, v in itertools.islice(data.items(), MAX_LOG_ITERABLE_SIZE)
            }
        if isinstance(data, str):
            if len(data) <= MAX_LOG_STRING_LENGTH:
                return data
            return f"{data[:MAX_LOG_STRING_LENGTH]}... <truncated>"
        if isinstance(data, (int, float, bool, type(None))):
            return data

        try:
            return json.loads(json.dumps(data))
        except (TypeError, OverflowError):
            return f"Can't be displayed in the log. Value of type: {str(type(data))}."

class GraphRunner:
    CANCELATION_CHECK_INTERVAL = 0.1

    def __init__(self, 
        graph: Graph,
        execution_environment: Dict,
        runner,
        title: str = "Blueprint execution"
    ):
        self.runner = runner
        self.graph = graph
        self.execution_environment = execution_environment
        self.run_id = execution_environment.get("blueprint_run_id", self._generate_run_id())
        execution_environment["blueprint_run_id"] = self.run_id
        self.status_logger = StatusLogger(self.graph, self.runner, self.run_id, title)

        self.queue = self.graph.get_start_nodes()
        self.futures: List[Future[GraphNode]] = []

    def run(self) -> Optional[Any]:
        if self.graph.status == "error":
            self.status_logger.log("Execution failed due to graph validation errors.", entry_type="error", exit="graph_validation_error")
            return None
        if not self.queue:
            raise WriterConfigurationError("No start nodes found in the blueprint.")

        with self.runner._get_executor() as executor:
            with self.runner.run_manager.register(self.run_id) as event:
                return self._execute(executor, event)

    def _execute(self, executor: ThreadPoolExecutor, abort_event: threading.Event) -> Optional[Any]:
        with use_journal_record_context(self.execution_environment, self.status_logger.title, self.graph) as journal_record:
            while self.queue or self.futures:
                while self.queue:
                    node: GraphNode = self.queue.pop(0)
                    if node.can_run() and node.outcome is None:
                        self.futures.append(node.run(self.execution_environment, self.runner, executor))

                self.status_logger.log("Executing...")
                done, _ = wait(self.futures, timeout=self.CANCELATION_CHECK_INTERVAL, return_when=FIRST_COMPLETED)
                if not done:
                    if abort_event.is_set():
                        self._cancel_all_jobs()
                        self.status_logger.log("Terminated.", entry_type="info", exit="aborted")
                        journal_record.set_result("stopped")
                        return None
                    else:
                        continue
                self.status_logger.log("Executing...")
                for future in done:
                    if future in self.futures:
                        self.futures.remove(future)
                    try:
                        result_node: GraphNode = future.result()
                    except BlueprintExecutionError as e:
                        self._cancel_all_jobs()
                        self.status_logger.log("Execution failed", entry_type="error", exit=str(e))
                        raise e
                    except BaseException as e:
                        abort_event.set()
                        self._cancel_all_jobs()
                        self.status_logger.log("Execution failed.", entry_type="error", exit=str(e))
                        raise BlueprintExecutionError(
                            f"Blueprint execution was stopped due to an error - {e.__class__.__name__}: {e}"
                        ) from e 
                    if result_node.outcome == "stopped":
                        continue 
                    if result_node.return_value is not None:
                        self._cancel_local_jobs()
                        self.status_logger.log(
                            f"Execution completed, node {result_node.id} returned value: {result_node.return_value}",
                            entry_type="info",
                            exit="return"
                        )
                        
                        journal_record.set_result("success")
                        return result_node.return_value
                    for output in result_node.outputs:
                        to_node_id = output.get("toNodeId")
                        next_node = self.graph.get_node(to_node_id)
                        if next_node:
                            self.queue.append(next_node)

            self.status_logger.log("Execution completed.", entry_type="info", exit="completed")
            journal_record.set_result("success")
            return None

    def _cancel_local_jobs(self):
        self.queue.clear()
        for node in self.graph.nodes:
            if node.outcome == "in_progress":
                node.status = "stopped"
        self.status_logger.log("Stopped")
        for future in self.futures:
            if not future.done():
                future.cancel()

    def _cancel_all_jobs(self):
        self._cancel_local_jobs()
        self.runner.cancel_blueprint_execution(self.run_id)

    def _generate_run_id(self):
        timestamp = str(int(time.time() * 1000))
        salt = os.urandom(8).hex()
        raw_id = f"{self.runner.session.session_id}_{timestamp}_{salt}"
        hashed_id = hashlib.sha256(raw_id.encode()).hexdigest()[:24]
        return hashed_id


def get_current_block() -> Optional[writer.blocks.base_block.BlueprintBlock]:
    return _current_block.get(None)


@contextmanager
def use_current_block(block: writer.blocks.base_block.BlueprintBlock):
    token = _current_block.set(block)
    yield
    _current_block.reset(token)



import contextlib
import json
import logging
from contextvars import ContextVar
from copy import deepcopy
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Dict, Literal, Optional

import writer.abstract
from writer.core import Config
from writer.keyvalue_storage import writer_kv_storage

if TYPE_CHECKING:
    from writer.blueprints import Graph, GraphNode
    from writer.core import Component


logger = logging.getLogger("journal")

JOURNAL_KEY_PREFIX = "wf-journal-"
INIT_LOGS_KEY_PREFIX = "wf-init-logs-"

class JournalRecord:
    def __init__(
        self,
        execution_environment: Dict,
        title: str,
        graph: "Graph"
    ):
        from writer import core_ui

        self.started_at = datetime.now(timezone.utc)
        self.instance_type = "editor" if Config.mode == "edit" else "agent"

        # Get blueprint_id from the parent of any node in the graph
        # All nodes in a blueprint share the same parent blueprint component
        self.blueprint_id = graph.nodes[0].component.parentId if graph.nodes else None

        self.execution_environment = execution_environment
        self.trigger = {
            "event": execution_environment.get("context", {}).get("event"),
            "payload": execution_environment.get("payload"),
            "component": {}
        }

        if self.trigger["event"] == "wf-run-blueprint":
            self.trigger["component"]["type"] = "blueprint"
            self.trigger["component"]["id"] = self.blueprint_id
            blueprint_component = core_ui.current_component_tree().get_component(self.trigger["component"]["id"])
            if blueprint_component is not None:
                self.trigger["component"]["title"] = blueprint_component.content.get("key")
        else:
            self.trigger["component"]["type"] = "block"
            component = graph.get_start_nodes()[0].component
            self.trigger["component"]["id"] = component.id
            self.trigger["component"]["title"] = self._get_block_info(component)["title"]

        if "API" in title:
            self.trigger["type"] = "API"
        elif "Cron" in title:
            self.trigger["type"] = "Cron"
        elif "UI" in title:
            self.trigger["type"] = "UI"
        else:
            self.trigger["type"] = "On demand"

        self.graph = graph
        self.block_outputs: Dict[str, Any] = {}
        for graph_node in self.graph.nodes:
            block_info = self._get_block_info(graph_node.component)
            self.block_outputs[graph_node.id] = {
                "component": {
                    "type": graph_node.component.type,
                    "id": graph_node.component.id,
                    "title": block_info["title"],
                    "category": block_info["category"]
                },
                "executions": []
            }

        self.is_runable = True
        self.result: Optional[Literal["success", "error", "stopped"]] = None

    def _get_block_info(self, component: "Component") -> Dict[str, str]:
        block_title = component.content.get("alias")
        component_definition = writer.abstract.templates.get(component.type)

        # If component has an alias, use it as title
        if block_title is not None:
            category = "Unknown category"
            if component_definition is not None:
                category = component_definition.writer.get("category", "Unknown category")
            return {
                "title": block_title,
                "category": category
            }

        # If no component definition found, return defaults
        if component_definition is None:
            return {
                "title": "Unknown block",
                "category": "Unknown category"
            }

        # Use component definition for both title and category
        return {
            "title": component_definition.writer.get("name", "Unknown block"),
            "category": component_definition.writer.get("category", "Unknown category")
        }

    def to_dict(self) -> Dict[str, Any]:
        block_outputs = deepcopy(self.block_outputs)
        for graph_node in self.graph.nodes:
            block_outputs[graph_node.id]["executions"].append(self.get_execution_data(graph_node))

        data = {
            "timestamp": self.started_at.isoformat(),
            "instanceType": self.instance_type,
            "blueprintId": self.blueprint_id,
            "trigger": self.trigger,
            "blockOutputs": block_outputs,
            "result": self.result,
        }
        sanitized_data = self._sanitize_data(data)
        return {
            **sanitized_data,
            "isRunable": self.is_runable,
        }

    def get_execution_data(self, graph_node: "GraphNode") -> Dict[str, Any]:
        execution_data: Dict[str, Any] = {
            "result": graph_node.result,
            "outcome": graph_node.outcome,
        }

        # Add timing information if available
        if graph_node.tool:
            if hasattr(graph_node.tool, 'started_at') and graph_node.tool.started_at >= 0:
                execution_data["startedAt"] = graph_node.tool.started_at
            if hasattr(graph_node.tool, 'execution_time_in_seconds') and graph_node.tool.execution_time_in_seconds >= 0:
                execution_data["executionTimeInSeconds"] = graph_node.tool.execution_time_in_seconds

            # Add captured logs if available
            if hasattr(graph_node.tool, 'captured_stdout') and graph_node.tool.captured_stdout:
                execution_data["stdout"] = graph_node.tool.captured_stdout
            if hasattr(graph_node.tool, 'captured_logs') and graph_node.tool.captured_logs:
                execution_data["logs"] = graph_node.tool.captured_logs

            # Add error message if available (contains the traceback for errors)
            if hasattr(graph_node.tool, 'message') and graph_node.tool.message:
                execution_data["message"] = graph_node.tool.message

        return execution_data

    def _sanitize_data(self, data):
        if data is None:
            return None

        if isinstance(data, list):
            return [self._sanitize_data(item) for item in data]
        if isinstance(data, dict):
            return {
                k: self._sanitize_data(v)
                for k, v in data.items()
            }
        if isinstance(data, (str, int, float, bool, type(None))):
            return data

        try:
            return json.loads(json.dumps(data))
        except (TypeError, OverflowError):
            self.is_runable = False
            return f"Can't be displayed in the Journal. Value of type: {str(type(data))}."

    def construct_key(self) -> str:
        return f"{JOURNAL_KEY_PREFIX}{self.instance_type[0]}-{int(self.started_at.timestamp() * 1000)}"
    
    def set_result(self, result: Literal["success", "error", "stopped"]) -> None:
        self.result = result

    def add_nested_execution(self, nested_record: "JournalRecord") -> None:
        for graph_node in nested_record.graph.nodes:
            if graph_node.id not in self.block_outputs:
                self.block_outputs[graph_node.id] = nested_record.block_outputs[graph_node.id]
            self.block_outputs[graph_node.id]["executions"].append(nested_record.get_execution_data(graph_node))
    
    def save(self) -> None:
        if "journal" not in Config.feature_flags or not writer_kv_storage.is_accessible():
            return
        data = self.to_dict()
        writer_kv_storage.save(self.construct_key(), data)


_parent_journal_record: ContextVar[Optional[JournalRecord]] = ContextVar("parent_journal_record", default=None)
_current_journal_record: ContextVar[Optional[JournalRecord]] = ContextVar("current_journal_record", default=None)

@contextlib.contextmanager
def use_journal_record_context(
    execution_environment: Dict,
    title: str,
    graph: "Graph"
):
    parent_record = _parent_journal_record.get()
    current_record = JournalRecord(execution_environment, title, graph)
    _current_journal_record.set(current_record)
    if parent_record is None:
        _parent_journal_record.set(current_record)

    try:
        yield current_record
    except BaseException as e:
        current_record.set_result("error")
        raise e
    finally:
        _current_journal_record.set(None)
        if parent_record is not None:
            parent_record.add_nested_execution(current_record)
        else:
            try:
                current_record.save()
            except Exception:
                logger.exception("Failed to save a Journal entry")
            _parent_journal_record.set(None)


def get_current_journal_record() -> Optional[JournalRecord]:
    return _current_journal_record.get()



import asyncio
import base64
import html
import importlib.util
import io
import json
import logging
import mimetypes
import os
import os.path
import pathlib
import socket
import tempfile
import textwrap
import time
import typing
from contextlib import asynccontextmanager, suppress
from importlib.machinery import ModuleSpec
from typing import (
    Any,
    AsyncGenerator,
    Callable,
    Dict,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)
from urllib.parse import urlsplit

import orjson
import uvicorn
from fastapi import FastAPI, File, HTTPException, Request, Response, UploadFile
from fastapi.responses import FileResponse, JSONResponse, StreamingResponse
from fastapi.routing import Mount
from fastapi.staticfiles import StaticFiles
from pydantic import ValidationError
from starlette.websockets import WebSocket, WebSocketDisconnect, WebSocketState

from writer import VERSION, abstract
from writer.ai import Graph
from writer.app_runner import AppRunner
from writer.ss_types import (
    AppProcessServerResponse,
    AutogenRequestBody,
    ComponentUpdateRequestPayload,
    DeleteDataRequestBody,
    EventResponsePayload,
    HashRequestPayload,
    HashRequestResponsePayload,
    InitRequestBody,
    InitResponseBodyEdit,
    InitResponseBodyRun,
    InitSessionRequestPayload,
    InitSessionResponsePayload,
    RetrieveDataRequestBody,
    RetrieveDataResponseBody,
    ServeMode,
    StateEnquiryResponsePayload,
    WriterEvent,
    WriterWebsocketIncoming,
    WriterWebsocketOutgoing,
)

if typing.TYPE_CHECKING:
    from .auth import Auth, Unauthorized

MAX_WEBSOCKET_MESSAGE_SIZE = 201 * 1024 * 1024
BLUEPRINT_API_EXECUTION_TIMEOUT_SECONDS = int(os.getenv("AGENT_BUILDER_BLUEPRINT_API_EXECUTION_TIMEOUT", "600"))
BLUEPRINT_API_RETRY_TIMEOUT = int(os.getenv("AGENT_BUILDER_BLUEPRINT_API_RETRY_TIMEOUT", "10000"))


class WriterState(typing.Protocol):
    app_runner: AppRunner
    writer_app: bool
    is_server_static_mounted: bool
    meta: Union[Dict[str, Any], Callable[[], Dict[str, Any]]]  # meta tags for SEO
    opengraph_tags: Union[
        Dict[str, Any], Callable[[], Dict[str, Any]]
    ]  # opengraph tags for social networks integration (facebook, discord)
    title: Union[str, Callable[[], str]]  # title of the page, default: "Writer Framework"


class WriterAsgi(typing.Protocol):
    state: WriterState


class WriterFastAPI(FastAPI, WriterAsgi):  # type: ignore
    pass


app: WriterFastAPI = cast(WriterFastAPI, None)


def get_asgi_app(
    user_app_path: str,
    serve_mode: ServeMode,
    enable_remote_edit: bool = False,
    enable_server_setup: bool = True,
    on_load: Optional[Callable] = None,
    on_shutdown: Optional[Callable] = None
) -> WriterFastAPI:
    """
    Builds an ASGI server that can be injected into another ASGI application
    or an asgi server like uvicorn

    >>> asgi_app = writer.serve.get_asgi_app("app1", "run")
    >>> uvicorn.run(asgi_app, host="0.0.0.0", port=5328)

    :param user_app_path: writer application path
    :param serve_mode: server mode (run, edit)
    :param enable_remote_edit: allow editing from the internet (by default, editing only works locally)
    :param enable_server_setup: enables fastapi setup hook on startup, server_setup.py
    :param on_load: callback called on loading
    :param on_shutdown: callback called at shutdown
    :return: ASGI Server
    """
    global app
    if serve_mode not in ["run", "edit"]:
        raise ValueError("""Invalid mode. Must be either "run" or "edit".""")

    _fix_mimetype()
    app_runner = AppRunner(user_app_path, serve_mode)
    pending_tasks: Set[asyncio.Task] = set()

    @asynccontextmanager
    async def lifespan(asgi_app: FastAPI):
        nonlocal app_runner

        app_runner.hook_to_running_event_loop()
        app_runner.load()

        if (
            on_load is not None
            and hasattr(asgi_app.state, "is_server_static_mounted")
            and asgi_app.state.is_server_static_mounted
        ):
            on_load()

        try:
            yield
        except asyncio.CancelledError:
            pass

        for pending_task in pending_tasks.copy():
            pending_task.cancel()
            try:
                await pending_task
            except asyncio.CancelledError:
                pass

        app_runner.shut_down()
        if on_shutdown is not None:
            on_shutdown()

    app = cast(WriterFastAPI, FastAPI(lifespan=lifespan))
    """
    Reuse the same pattern to give variable to FastAPI application
    than `app.state.is_server_static_mounted` already use in Writer Framework.
    """
    app.state.writer_app = True
    app.state.app_runner = app_runner

    def _get_extension_paths() -> List[str]:
        extensions_path = pathlib.Path(user_app_path) / "extensions"
        if not extensions_path.exists():
            return []
        filtered_files = [
            f
            for f in extensions_path.rglob("*")
            if f.suffix.lower() in (".js", ".css") and f.is_file()
        ]
        relative_paths = [f.relative_to(extensions_path).as_posix() for f in filtered_files]
        return relative_paths

    cached_extension_paths = _get_extension_paths()

    def _check_origin_header(origin_header: Optional[str]) -> bool:
        if serve_mode not in ("edit") or enable_remote_edit is True:
            return True
        if origin_header is None:
            return False
        hostname = urlsplit(origin_header).hostname
        if hostname in ("127.0.0.1", "localhost"):
            return True
        return False

    # Init

    def _get_run_starter_pack(payload: InitSessionResponsePayload):
        return InitResponseBodyRun(
            mode="run",
            sessionId=payload.sessionId,
            userState=payload.userState,
            mail=payload.mail,
            components=payload.components,
            userFunctions=payload.userFunctions,
            extensionPaths=cached_extension_paths,
            featureFlags=payload.featureFlags,
            abstractTemplates=abstract.templates,
            writerApplication=payload.writerApplication,
        )

    def _get_edit_starter_pack(payload: InitSessionResponsePayload):
        run_code: Optional[str] = app_runner.run_code

        return InitResponseBodyEdit(
            mode="edit",
            sessionId=payload.sessionId,
            userState=payload.userState,
            mail=payload.mail,
            components=payload.components,
            userFunctions=payload.userFunctions,
            runCode=run_code,
            sourceFiles=app_runner.source_files,
            extensionPaths=cached_extension_paths,
            featureFlags=payload.featureFlags,
            abstractTemplates=abstract.templates,
            writerApplication=payload.writerApplication,
        )

    @app.get("/api/health")
    async def health():
        app_runner = app.state.app_runner
        
        # Check user app process
        if app_runner.app_process is None or not app_runner.app_process.is_alive():
            return JSONResponse(
                status_code=503,
                content={"status": "error", "message": "User app process is not running"}
            )
        
        # Check project saver process (only in edit mode)
        if app_runner.mode == "edit":
            project_saver = app_runner.wf_project_context.write_files_async_process
            if project_saver is None or not project_saver.is_alive():
                return JSONResponse(
                    status_code=503,
                    content={"status": "error", "message": "Project saver process is not running"}
                )
        
        return {"status": "ok"}

    @app.get("/api/export")
    async def export_zip():
        if serve_mode != "edit":
            raise HTTPException(status_code=403, detail="Invalid mode.")
        exported_zip_stream = app_runner.export_zip()
        return StreamingResponse(
            exported_zip_stream,
            media_type="application/x-zip-compressed",
            headers={
                "Content-Disposition": "attachment; filename=exported_agent.zip"
            }
        )

    @app.post("/api/import")
    async def import_zip(file: UploadFile = File(...)):
        if serve_mode != "edit":
            raise HTTPException(status_code=403, detail="Invalid mode.")
        if not file.filename or not file.filename.endswith(".zip"):
            raise HTTPException(status_code=400, detail="Only .zip files are supported.")

        MAX_FILE_SIZE = 200 * 1024 * 1024

        try:
            with tempfile.NamedTemporaryFile(delete=False) as tmp:
                # Stream file to disk to avoid memory issues
                size = 0
                while chunk := await file.read(8192):
                    size += len(chunk)
                    if size > MAX_FILE_SIZE:
                        tmp.close()
                        os.unlink(tmp.name)
                        raise HTTPException(status_code=413, detail=f"File too large. Max file size: {MAX_FILE_SIZE}")
                    tmp.write(chunk)
                tmp_path = tmp.name
            await app_runner.import_zip(tmp_path)
            os.remove(tmp_path)
        except ValueError:
            raise HTTPException(status_code=400, detail="Invalid upload.")

    @app.post("/api/autogen")
    async def autogen(requestBody: AutogenRequestBody, request: Request):
        import writer.autogen
        agent_token_header = request.headers.get('x-agent-token')

        return writer.autogen.generate_blueprint(
            requestBody.description,
            agent_token_header
            )

    @app.post("/api/data/retrieve")
    async def retrieve_data(requestBody: RetrieveDataRequestBody) -> RetrieveDataResponseBody:
        from writer.keyvalue_storage import writer_kv_storage

        all_keys = writer_kv_storage.get_data_keys()

        keys_to_fetch = []
        for key in all_keys:
            if key in requestBody.skip_keys:
                continue
            if requestBody.key_contains and requestBody.key_contains not in key:
                continue
            keys_to_fetch.append(key)
        
        async def fetch_value(key: str):
            return key, await asyncio.to_thread(writer_kv_storage.get, key, "data")

        kv_pairs = await asyncio.gather(*(fetch_value(key) for key in keys_to_fetch))

        return RetrieveDataResponseBody(result={k: v["data"] for k, v in kv_pairs})

    @app.post("/api/data/delete")
    async def delete_data(requestBody: DeleteDataRequestBody) -> None:
        from writer.keyvalue_storage import writer_kv_storage

        async def delete_key(key: str):
            return key, await asyncio.to_thread(writer_kv_storage.delete, key)

        await asyncio.gather(*(delete_key(key) for key in requestBody.keys))

        return None

    @app.post("/api/init")
    async def init(
        initBody: InitRequestBody, request: Request, response: Response
    ) -> Union[InitResponseBodyRun, InitResponseBodyEdit]:
        """
        Handles session init and provides a "starter pack" to the frontend.
        """

        origin_header = request.headers.get("origin")
        if not _check_origin_header(origin_header):
            wrong_origin_message = "A session request with origin %s was rejected. For security reasons, only local origins are allowed in edit mode. "
            wrong_origin_message += "To circumvent this protection, use the --enable-remote-edit flag if running via command line."
            logging.error(wrong_origin_message, origin_header)
            raise HTTPException(
                status_code=403, detail="Incorrect origin. Only local origins are allowed."
            )

        session_id = request.cookies.get("session")
        if session_id is not None:
            initBody.proposedSessionId = session_id

        app_response = await app_runner.init_session(
            InitSessionRequestPayload(
                cookies=dict(request.cookies),
                headers=dict(request.headers),
                proposedSessionId=initBody.proposedSessionId,
            )
        )

        status = app_response.status

        """
        Deletes the session cookie that was set by 
        authentication when it is present.
        """
        if session_id is not None:
            response.delete_cookie("session")

        if status == "error" or app_response.payload is None:
            raise HTTPException(status_code=403, detail="Session rejected.")

        if serve_mode == "run":
            return _get_run_starter_pack(app_response.payload)

        if serve_mode == "edit":
            return _get_edit_starter_pack(app_response.payload)

    # Jobs

    async def _get_payload_as_json(request: Request):
        payload = None
        body = await request.body()
        if not body:
            return None
        try:
            payload = await request.json()
        except json.JSONDecodeError:
            raise HTTPException(status_code=400, detail="Cannot parse the payload.")
        return payload
    
    def is_blueprint_triggerable(app_runner: AppRunner, blueprint_id: str) -> bool:
        """Check if blueprint has at least one supported trigger (API or Cron)."""
        if not app_runner.bmc_components:
            return False
        supported_triggers = ("blueprints_apitrigger", "blueprints_crontrigger")
        return any(
            comp["type"] in supported_triggers and comp.get("parentId") == blueprint_id
            for comp in app_runner.bmc_components.values()
        )

    @app.get("/private/api/blueprints")
    async def get_blueprints(request: Request):
        """
        Returns a list of blueprints available in the agent.
        """
        if not app_runner.bmc_components:
            return JSONResponse(content=[])

        blueprints = [
            {
                "id": comp["id"],
                "key": comp.get("content", {}).get("key")
            }
            for comp in app_runner.bmc_components.values()
            if comp["type"] == "blueprints_blueprint"
            and is_blueprint_triggerable(app_runner, comp["id"])
        ]

        return JSONResponse(content=blueprints)

    @app.get("/private/api/cron-triggers")
    async def get_cron_triggers(request: Request):
        """
        Returns a list of Cron Trigger blocks.
        """
        if not app_runner.bmc_components:
            return JSONResponse(content=[], status_code=200)
        
        definition = abstract.templates["blueprints_crontrigger"].writer

        cron_triggers = [
            {
                "id": comp.get("id"),
                "blueprint_id": comp.get("parentId"),
                "name": comp.get("content", {}).get("alias") or definition["name"],
                "cron_expression": comp.get("content", {}).get("cronExpression", ""),
                "timezone": comp.get("content", {}).get("timezone", "UTC"),
            }
            for comp in app_runner.bmc_components.values()
            if comp.get("type") == "blueprints_crontrigger"
        ]

        return JSONResponse(content=cron_triggers, status_code=200)

    @app.post("/private/api/blueprint/{blueprint_id}")
    async def create_blueprint_job(blueprint_id: str, request: Request, response: Response, branch_id: Optional[str] = None):
        # Keep-alive interval for SSE streaming
        KEEPALIVE_INTERVAL = 15
        payload = await _get_payload_as_json(request)

        # --- Session initialization ---

        async def init_session_and_validate(
            app_runner: AppRunner,
            cookies: Dict[str, Any],
            headers: Dict[str, Any],
        ) -> str:
            # Initialize session with passed cookies/headers
            sess_resp = await app_runner.init_session(InitSessionRequestPayload(
                cookies=cookies, headers=headers, proposedSessionId=None
            ))
            if not sess_resp or not sess_resp.payload:
                raise RuntimeError("Cannot initialize session.")
            sid = sess_resp.payload.sessionId
            if not await app_runner.check_session(sid):
                raise RuntimeError("Cannot initialize session.")
            return sid

        # --- Blueprint discovery logic ---

        def check_blueprint(app_runner: AppRunner, blueprint_id: str) -> bool:
            # Locate blueprint component by its key
            if not app_runner.bmc_components:
                return False
            return blueprint_id in app_runner.bmc_components

        # --- Result serialization (recursive) ---

        def serialize_result(data: Any) -> Any:
            # Convert blueprint output into JSON-serializable structure
            if isinstance(data, (str, int, float, bool, type(None))):
                return data
            if isinstance(data, list):
                return [serialize_result(item) for item in data]
            if isinstance(data, dict):
                return {k: serialize_result(v) for k, v in data.items()}
            try:
                return json.loads(json.dumps(data))
            except (TypeError, OverflowError):
                return f"Can't be displayed. Value of type: {type(data)}."

        # --- SSE formatting utilities ---

        async def format_event(event_type: str, data: Dict[str, Any]) -> str:
            # Format a proper Server-Sent Event chunk
            return f"event: {event_type}\ndata: {json.dumps(data)}\n\n"

        async def format_keepalive() -> str:
            # Send a SSE comment line as keep-alive (spec compliant)
            return ": keep-alive\n\n"

        # --- The main worker logic that produces events ---

        async def event_logic(queue: asyncio.Queue):
            try:
                await queue.put(await format_event("status", {"status": "in progress", "created_at": int(time.time())}))
                await queue.put(await format_event("status", {"status": "initializing", "msg": "Initializing session..."}))

                # Validate session & credentials
                session_id = await init_session_and_validate(
                    app_runner, dict(request.cookies), dict(request.headers)
                )

                await queue.put(await format_event("status", {"status": "validating", "msg": "Validating blueprint..."}))

                if not app_runner.bmc_components:
                    raise RuntimeError("No blueprints defined in the agent.")

                blueprint_exists = check_blueprint(app_runner, blueprint_id)
                if not blueprint_exists:
                    await queue.put(await format_event("error", {
                        "msg": f"Blueprint '{blueprint_id}' was not found.",
                        "finished_at": int(time.time())
                    }))
                    return

                if not branch_id and not is_blueprint_triggerable(app_runner, blueprint_id):
                    await queue.put(await format_event("error", {
                        "msg": f"Blueprint '{blueprint_id}' lacks a supported trigger (API or Cron).",
                        "finished_at": int(time.time())
                    }))
                    return

                if branch_id:
                    block = app_runner.bmc_components.get(branch_id)
                    if not block:
                        await queue.put(await format_event("error", {
                            "msg": f"Block '{branch_id}' was not found.",
                            "finished_at": int(time.time())
                        }))
                        return

                await queue.put(await format_event("status", {"status": "executing", "msg": (f"Executing branch: {branch_id}..." if branch_id else f"Executing blueprint: {blueprint_id}...")}))

                task = asyncio.create_task(
                    app_runner.handle_event(
                        session_id,
                        WriterEvent(
                            type="wf-run-blueprint-via-api",
                            isSafe=True,
                            handler="run_blueprint_via_api",
                            payload={
                                "blueprint_id": blueprint_id,
                                "branch_id": branch_id,
                                **(payload or {})
                            },
                        )
                    )
                )

                await queue.put(await format_event("status", {"status": "running", "msg": ("Branch is running. Awaiting output..." if branch_id else "Blueprint is running. Awaiting output...")}))

                # Await blueprint execution with timeout protection
                apsr = await asyncio.wait_for(task, timeout=BLUEPRINT_API_EXECUTION_TIMEOUT_SECONDS)

                await queue.put(await format_event("status", {"status": "processing", "msg": "Processing blueprint result..."}))

                if not apsr or apsr.status != "ok":
                    raise RuntimeError("Blueprint execution failed.")

                if apsr.payload and apsr.payload.result:
                    task_status = apsr.payload.result.get("ok", False)
                    result = serialize_result(
                        apsr.payload.result.get("result")
                    )
                else:
                    task_status = False
                    result = "No result returned from blueprint execution."

                if not task_status:
                    await queue.put(await format_event("error", {
                        "msg": result,
                        "finished_at": int(time.time())
                    }))
                else:
                    await queue.put(await format_event("artifact", {
                        "artifact": result,
                        "finished_at": int(time.time())
                    }))

            except Exception as e:
                # Bubble up any unexpected error as 'error' SSE event
                await queue.put(await format_event("error", {
                    "msg": f"Agent Builder internal error: {str(e)}",
                    "finished_at": int(time.time())
                }))
            finally:
                # Always mark stream completion for consumer
                await queue.put("data: [DONE]\n\n")

        # --- The streaming loop that multiplexes events and keep-alives ---

        async def merged_stream() -> AsyncGenerator[str, None]:
            # Type annotation required by mypy
            queue: asyncio.Queue = asyncio.Queue()
            producer_task = asyncio.create_task(event_logic(queue))

            yield f"retry: {BLUEPRINT_API_RETRY_TIMEOUT}\n\n"

            try:
                while True:
                    try:
                        result = await asyncio.wait_for(
                            queue.get(),
                            timeout=KEEPALIVE_INTERVAL
                            )
                        if result == "data: [DONE]\n\n":
                            return
                        yield result
                    except asyncio.TimeoutError:
                        yield await format_keepalive()
                    except asyncio.CancelledError:
                        # Client disconnected, break streaming loop
                        break
            finally:
                # Always cancel producer to prevent orphaned task
                producer_task.cancel()
                with suppress(asyncio.CancelledError):
                    await producer_task

        return StreamingResponse(
            merged_stream(),
            media_type="text/event-stream",
            headers={
                "Cache-Control": "no-cache",
                "Connection": "keep-alive",
                "Access-Control-Allow-Origin": "*",
                "Access-Control-Allow-Headers": "Cache-Control",
            },
        )

    # Streaming

    async def _send_json_or_queue(session_id: str, data: Any, websocket: WebSocket):
        try:
            binary_data = orjson.dumps(data)
            await websocket.send_bytes(binary_data)
        except (RuntimeError, WebSocketDisconnect):
            await app_runner.queue_message(session_id, data)

    async def _stream_session_init(websocket: WebSocket):
        """
        Waits for the client to provide a session id to initialise the stream.
        Returns the session id received.
        """

        session_id = None
        while session_id is None:
            req_message_raw = await websocket.receive_json()

            try:
                req_message = WriterWebsocketIncoming.model_validate(req_message_raw)
            except ValidationError:
                logging.error("Incorrect incoming request.")
                return

            if req_message.type == "streamInit" and req_message.payload is not None:
                session_id = req_message.payload.get("sessionId")
        return session_id

    async def _stream_incoming_requests(websocket: WebSocket, session_id: str):
        """
        Handles incoming requests from client.
        """

        try:
            while True:
                req_message_raw = await websocket.receive_json()

                try:
                    req_message = WriterWebsocketIncoming.model_validate(req_message_raw)
                except ValidationError:
                    logging.error("Incorrect incoming request.")
                    break

                is_session_ok = await app_runner.check_session(session_id)
                if not is_session_ok:
                    break

                new_task = None

                if req_message.type == "event":
                    new_task = asyncio.create_task(
                        _handle_incoming_event(websocket, session_id, req_message)
                    )
                elif req_message.type == "keepAlive":
                    new_task = asyncio.create_task(
                        _handle_keep_alive_message(websocket, session_id, req_message)
                    )
                elif req_message.type == "stateEnquiry":
                    new_task = asyncio.create_task(
                        _handle_state_enquiry_message(websocket, session_id, req_message)
                    )
                elif serve_mode == "edit" and req_message.type == "hashRequest":
                    new_task = asyncio.create_task(
                        _handle_hash_request(websocket, session_id, req_message)
                    )
                elif serve_mode == "edit":
                    new_task = asyncio.create_task(
                        _handle_incoming_edit_message(websocket, session_id, req_message)
                    )

                if new_task:
                    pending_tasks.add(new_task)
                    new_task.add_done_callback(pending_tasks.discard)
        except WebSocketDisconnect:
            return
        except asyncio.CancelledError:
            raise

    async def _handle_incoming_event(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType=f"{req_message.type}Response",
            trackingId=req_message.trackingId,
            payload=None,
        )

        # Allows for global events if in edit mode (such as "Run blueprint" for previewing a blueprint)

        is_safe = serve_mode == "edit"
        res_payload: Optional[Dict[str, Any]] = None
        apsr: Optional[AppProcessServerResponse] = None
        apsr = await app_runner.handle_event(
            session_id,
            WriterEvent(
                type=req_message.payload.get("type"),
                handler=req_message.payload.get("handler"),
                isSafe=is_safe,
                instancePath=req_message.payload.get("instancePath"),
                payload=req_message.payload.get("payload"),
            ),
        )
        if apsr is not None and apsr.payload is not None:
            res_payload = typing.cast(EventResponsePayload, apsr.payload).model_dump()
        if res_payload is not None:
            response.payload = res_payload
        await _send_json_or_queue(session_id, response.model_dump(), websocket)

    async def _handle_incoming_edit_message(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType=f"{req_message.type}Response",
            trackingId=req_message.trackingId,
            payload=None,
        )
        if req_message.type == "componentUpdate":
            await app_runner.update_components(
                session_id,
                ComponentUpdateRequestPayload(components=req_message.payload["components"]),
            )
            await app_runner.queue_announcement_async(
                "componentUpdate", req_message.payload["components"], session_id
            )
        elif req_message.type == "collaborationPing":
            await app_runner.queue_announcement_async(
                "collaborationUpdate", req_message.payload, exclude_session_id=session_id
            )
        elif req_message.type == "codeSaveRequest":
            app_runner.save_code(
                session_id, req_message.payload["code"], req_message.payload["path"]
            )
        elif req_message.type == "codeUpdate":
            app_runner.update_code(session_id, req_message.payload["code"])
        elif req_message.type == "loadSourceFile":
            path = os.path.join(*req_message.payload["path"])
            try:
                response.payload = {"content": app_runner.load_persisted_script(path)}
            except FileNotFoundError as error:
                logging.warning(f"could not load script at {path}", error)
                response.payload = {"error": str(error)}
        elif req_message.type == "createSourceFile":
            path = os.path.join(*req_message.payload["path"])
            try:
                app_runner.create_persisted_script(path)
            except Exception as error:
                response.payload = {"error": str(error)}
        elif req_message.type == "deleteSourceFile":
            path = os.path.join(*req_message.payload["path"])
            try:
                app_runner.delete_persisted_script(path)
            except Exception as error:
                response.payload = {"error": str(error)}
        elif req_message.type == "renameSourceFile":
            from_path = os.path.join(*req_message.payload["from"])
            to_path = os.path.join(*req_message.payload["to"])
            try:
                app_runner.rename_persisted_script(from_path, to_path)
            except Exception as error:
                response.payload = {"error": str(error)}
        elif req_message.type == "listResources":
            res = await app_runner.list_resources(session_id, req_message.payload["resource_type"])
            response.payload = res.payload
        elif req_message.type == "uploadSourceFile":
            path = os.path.join(*req_message.payload["path"])

            try:
                content = base64.b64decode(req_message.payload["content"])
                app_runner.create_persisted_script(path, content)
                response.payload = {"sourceFiles": app_runner.source_files}
            except Exception as error:
                response.payload = {"error": str(error)}
        elif req_message.type == "writerVaultUpdate":
            await app_runner.writer_vault_refresh(session_id)

        await _send_json_or_queue(session_id, response.model_dump(), websocket)

    async def _handle_keep_alive_message(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType="keepAliveResponse", trackingId=req_message.trackingId, payload=None
        )
        await _send_json_or_queue(session_id, response.model_dump(), websocket)

    async def _handle_state_enquiry_message(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType=f"{req_message.type}Response",
            trackingId=req_message.trackingId,
            payload=None,
        )
        res_payload: Optional[Dict[str, Any]] = None
        apsr: Optional[AppProcessServerResponse] = None
        apsr = await app_runner.handle_state_enquiry(session_id)
        if apsr is not None and apsr.payload is not None:
            res_payload = typing.cast(StateEnquiryResponsePayload, apsr.payload).model_dump()
        if res_payload is not None:
            response.payload = res_payload
        await _send_json_or_queue(session_id, response.model_dump(), websocket)

    async def _handle_hash_request(
        websocket: WebSocket, session_id: str, req_message: WriterWebsocketIncoming
    ):
        response = WriterWebsocketOutgoing(
            messageType=f"{req_message.type}Response",
            trackingId=req_message.trackingId,
            payload=None,
        )
        apsr: Optional[AppProcessServerResponse] = None
        apsr = await app_runner.handle_hash_request(
            session_id, HashRequestPayload(message=req_message.payload.get("message", ""))
        )
        if apsr is not None and apsr.payload is not None:
            response.payload = typing.cast(HashRequestResponsePayload, apsr.payload).model_dump()
        await _send_json_or_queue(session_id, response.model_dump(), websocket)

    async def _stream_outgoing_announcements(websocket: WebSocket, session_id: str):
        """
        Handles outgoing communications to the client (announcements).
        """

        WEBSOCKET_CODE_UPDATE_CODE = 4001
        session_queue: asyncio.Queue = asyncio.Queue()
        app_runner.announcement_queues[session_id] = session_queue

        try:
            while True:
                announcement_data = await session_queue.get()
                announcement = WriterWebsocketOutgoing(
                    messageType="announcement", trackingId=-1, payload=announcement_data
                )
                if websocket.application_state == WebSocketState.CONNECTED:
                    await websocket.send_json(announcement.dict())
                if announcement_data.get("type") == "codeUpdate":
                    await websocket.close(WEBSOCKET_CODE_UPDATE_CODE, "Code update.")
                    return
        except WebSocketDisconnect:
            pass
        except asyncio.CancelledError:
            raise
        finally:
            if app_runner.announcement_queues.get(session_id) is None:
                return
            del app_runner.announcement_queues[session_id]

    @app.websocket("/api/stream")
    async def stream(websocket: WebSocket):
        """Initialises incoming and outgoing communications on the stream."""

        await websocket.accept()

        origin_header = websocket.headers.get("origin")
        if not _check_origin_header(origin_header):
            await websocket.close(code=1008)
            return

        try:
            session_id = await _stream_session_init(websocket)
        except WebSocketDisconnect:
            return

        is_session_ok = await app_runner.check_session(session_id)
        if not is_session_ok:
            await websocket.close(code=1008)  # Invalid permissions
            return
        
        try:
            queued_messages = await app_runner.retrieve_messages(session_id)
            for message in queued_messages:
                await websocket.send_json(message)
            await app_runner.clear_messages(session_id)
        except (WebSocketDisconnect, RuntimeError):
            return

        task1 = asyncio.create_task(_stream_incoming_requests(websocket, session_id))
        task2 = asyncio.create_task(_stream_outgoing_announcements(websocket, session_id))

        try:
            await asyncio.wait((task1, task2), return_when=asyncio.FIRST_COMPLETED)
            await asyncio.sleep(1)
            task1.cancel()
            task2.cancel()
            await task1
            await task2
        except asyncio.CancelledError:
            pass

    # Mount static paths

    user_app_static_path = pathlib.Path(user_app_path) / "static"
    if user_app_static_path.exists():
        app.mount("/static", StaticFiles(directory=str(user_app_static_path)), name="user_static")

    user_app_extensions_path = pathlib.Path(user_app_path) / "extensions"
    if user_app_extensions_path.exists():
        app.mount(
            "/extensions", StaticFiles(directory=str(user_app_extensions_path)), name="extensions"
        )

    server_path = pathlib.Path(__file__)
    server_static_path = server_path.parent / "static"
    if server_static_path.exists():
        _mount_render_index_html(app, server_static_path)
        _mount_server_static_path(app, server_static_path)
        app.state.is_server_static_mounted = True
    else:
        logging.error(
            textwrap.dedent(
                """\
                \x1b[31;20mError: Failed to acquire server static path. Writer Framework may not be properly built.

                To resolve this issue, try the following steps:
                1. Run the 'npm run build' script in the 'ui' directory and then restart the app.
                2. Alternatively, launch a UI instance by running 'npm run dev' in the 'ui' directory.

                Please refer to the CONTRIBUTING.md for detailed instructions.\x1b[0m"""
            )
        )

    # Return
    if enable_server_setup is True:
        _execute_server_setup_hook(user_app_path)

    return app


def print_init_message():
    print(f"""               
                                                               
                   &@@@@@@@@@@     ,@@@@@@@@@@*     @@@@@@@@@@                  
                   .@@@@@@@@@@(     &@@@@@@@@@@     *@@@@@@@@*                  
                    %@@@@@@@@@@     .@@@@@@@@@@(     @@@@@@@@                   
                     @@@@@@@@@@&     #@@@@@@@@@@      @@@@@@                    
                     ,@@@@@@@@@@,     @@@@@@@@@@@     (@@@@(                    
                      &@@@@@@@@@@     .@@@@@@@@@@*     @@@@                     
                       @@@@@@@@@@%     %@@@@@@@@@@     .@@,                     
                       (@@@@@@@@@@.     @@@@@@@@@@%     %&                      
                        @@@@@@@@@@&     (@@@@@@@@@@      .                      
                        *@@@@@@@@@@,     @@@@@@@@@@&                            
                         @@@@@@@@@@@     ,@@@@@@@@@@*                           
                          @@@@@@@@@@#     %@@@@@@@@@@ 


WRITER FRAMEWORK v{VERSION}""")


def print_route_message(run_name: str, port: int, host: str):
    GREEN_TOKEN = "\033[92m"
    END_TOKEN = "\033[0m"

    print(
        f"{run_name} is available at:{END_TOKEN}{GREEN_TOKEN} http://{host}:{port}{END_TOKEN}",
        flush=True,
    )


def register_auth(
    auth: "Auth",
    callback: Optional[Callable[[Request, str, dict], None]] = None,
    unauthorized_action: Optional[Callable[[Request, "Unauthorized"], Response]] = None,
):
    auth.register(app, callback=callback, unauthorized_action=unauthorized_action)


def serve(
    app_path: str,
    mode: ServeMode,
    port: Optional[int],
    host,
    enable_remote_edit=False,
    enable_server_setup=False
):
    """Initialises the web server."""

    print_init_message()

    def on_load():
        run_name = "Builder" if mode == "edit" else "App"
        print_route_message(run_name, port, host)

    """
    Loading of the server_setup.py is active by default 
    when Writer Framework is launched with the run command.
    """
    if port is None:
        mode_allowed_ports = {"run": (3005, 3099), "edit": (4005, 4099)}

        port = _next_localhost_available_port(mode_allowed_ports[mode])

    enable_server_setup = mode == "run" or enable_server_setup
    app = get_asgi_app(
        app_path,
        mode,
        enable_remote_edit,
        on_load=on_load,
        enable_server_setup=enable_server_setup
    )
    log_level = "warning"
    uvicorn.run(
        app, host=host, port=port, log_level=log_level, ws_max_size=MAX_WEBSOCKET_MESSAGE_SIZE
    )


@asynccontextmanager
async def lifespan(app: FastAPI):
    """
    This feature supports launching multiple Writer Framework applications simultaneously.

    >>> import uvicorn
    >>> import writer.serve
    >>> from fastapi import FastAPI, Response
    >>>
    >>> root_asgi_app = FastAPI(lifespan=writer.serve.lifespan)
    >>>
    >>> sub_asgi_app_1 = writer.serve.get_asgi_app("../app1", "run")
    >>> sub_asgi_app_2 = writer.serve.get_asgi_app("../app2", "run")
    >>>
    >>> uvicorn.run(root_asgi_app, ws_max_size=writer.serve.MAX_WEBSOCKET_MESSAGE_SIZE)

    Writer Framework uses lifespan to start an application server (app_runner) per
    application.
    """
    writer_lifespans = []
    for route in app.routes:
        if isinstance(route, Mount) and isinstance(route.app, FastAPI):
            if hasattr(route.app.state, "writer_app"):
                ctx = route.app.router.lifespan_context
                writer_lifespans.append(ctx)

    async with _lifespan_invoke(writer_lifespans, app):
        yield


def configure_webpage_metadata(
    title: Union[str, Callable[[], str]] = "Writer Framework",
    meta: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
    opengraph_tags: Optional[Union[Dict[str, Any], Callable[[], Dict[str, Any]]]] = None,
):
    """
    Configures the page header for SEO and social networks from `server_setup` module.

    >>> writer.serve.configure_webpage_metadata(
    >>>     title="my App",
    >>>     meta={
    >>>         "description": "my amazing app",
    >>>         "keywords": "WF, Amazing, AI App",
    >>>         "author": "Amazing company"
    >>>     }
    >>>)

    Meta will accept description, keywords, author. Other meta tags as view port won't be supported.

    Settings accept functions to adapt content based on application data.

    >>> def generated_title():
    >>>     return "My App" # load title using info from database

    >>> def generated_meta_tags():
    >>>     {
    >>>         "description": "my amazing app",
    >>>         "keywords": "WF, Amazing, AI App",
    >>>         "author": "Amazing company"
    >>>     }

    >>> writer.serve.configure_webpage_metadata(
    >>>     title=generated_title
    >>>     meta=generated_meta_tags
    >>> )

    OpenGraph tags are used by social networks to display information about the page. WF support them.

    >>> writer.serve.configure_webpage_metadata(
    >>>     title=generated_title
    >>>     opengraph_tags= {
    >>>         "og:title": "My App",
    >>>         "og:description": "My amazing app",
    >>>         "og:image": "https://myapp.com/logo.png",
    >>>         "og:url": "https://myapp.com"
    >>>     }
    >>> )

    >>> def generated_opengraph_tags():
    >>>     return {
    >>>         "og:title": "My App",
    >>>         "og:description": "My amazing app",
    >>>     }

    >>> writer.serve.configure_webpage_metadata(
    >>>     title=generated_title
    >>>     opengraph_tags= generated_opengraph_tags
    >>> )

    ---

    WF replaces the placeholders <!-- {{ meta }} -->, <!-- {{ opengraph_tags }} -->
    and the <title>Writer framework<title> tag in the index.html file.

    :param title: The title of the page. Default: "Writer Framework"
    :param meta: A list of meta tags. Default: {}
    :param opengraph_tags: A dictionary of OpenGraph tags. Default: {}
    """
    app.state.title = title
    app.state.meta = meta if meta is not None else {}
    app.state.opengraph_tags = opengraph_tags if opengraph_tags is not None else {}


@asynccontextmanager
async def _lifespan_invoke(context: list, app: FastAPI):
    """
    Helper to run multiple lifespans in cascade.

    Running

    >>> _lifespan_invoke([app1.router.lifespan_context, app2.router.lifespan_context], app)

    is equivalent to

    >>> @asynccontextmanager
    >>> async def lifespan_context(app: FastAPI):
    >>>   async with app1.router.lifespan_context(app):
    >>>     async with app2.router.lifespan_context(app):
    >>>       yield
    """
    ctx = context.pop(0)
    async with ctx(app):
        if len(context) > 0:
            async with _lifespan_invoke(context, app):
                yield
        else:
            yield


def _fix_mimetype():
    """
    Fixes mimetypes for .js files. This is needed for the webserver to serve .js files correctly.
    """
    js_mimetype = mimetypes.guess_type("myfile.js")[0]
    if js_mimetype[0] != "text/javascript":
        mimetypes.add_type("text/javascript", ".js")


def _mount_server_static_path(app: FastAPI, server_static_path: pathlib.Path) -> None:
    """
    Unitarily declares the files and folders present in "/static" directory of source code.

    We avoid the general declaration as below. This declaration limit the ability of a developper to
    declare it's own route.

    >>> asgi_app.mount("/", StaticFiles(directory=str(server_static_path), html=True), name="server_static")

    Writer Framework routes remain priority. A developer cannot come and overload them.
    """
    for f in wf_root_static_assets():
        if f.is_file():
            app.get(f"/{f.name}")(lambda: FileResponse(f))
        if f.is_dir():
            app.mount(f"/{f.name}", StaticFiles(directory=f), name=f"server_static_{f}")


def _mount_render_index_html(app: FastAPI, server_static_path: pathlib.Path):
    """
    Serves the main page with the title that has been configured.

    :param app:
    :param server_static_path:
    :return:
    """

    def _render_index_html():
        with io.open(server_static_path.joinpath("index.html"), "r", encoding="utf-8") as f:
            index_html = f.read()
            if hasattr(app.state, "title"):
                index_html = index_html.replace(
                    "<title>Writer Framework</title>",
                    f"<title>{html.escape(app.state.title)}</title>",
                )

            if hasattr(app.state, "meta"):
                meta = app.state.meta() if callable(app.state.meta) else app.state.meta
                meta_tags = "\n".join(
                    [f'<meta name="{k}" content="{html.escape(v)}">' for k, v in meta.items()]
                )
                index_html = index_html.replace("<!-- {{ meta }} -->", meta_tags)
            else:
                index_html = index_html.replace("<!-- {{ meta }} -->", "")

            if hasattr(app.state, "opengraph_tags"):
                opengraph_tags = (
                    app.state.opengraph_tags()
                    if callable(app.state.opengraph_tags)
                    else app.state.opengraph_tags
                )
                opengraph_tags = "\n".join(
                    [
                        f'<meta property="{k}" content="{html.escape(v)}">'
                        for k, v in opengraph_tags.items()
                    ]
                )
                index_html = index_html.replace("<!-- {{ opengraph_tags }} -->", opengraph_tags)
            else:
                index_html = index_html.replace("<!-- {{ opengraph_tags }} -->", "")

        return Response(content=index_html, media_type="text/html")

    return app.get("/")(_render_index_html)


def app_runner(asgi_app: WriterFastAPI) -> AppRunner:
    return asgi_app.state.app_runner


def wf_root_static_assets() -> List[pathlib.Path]:
    """
    Lists the root writer Framework static assets. Some of them are files, some other are directories.

    >>> for f in wf_root_static_assets()
    >>>     print(f"{f.name}")
    >>>     # favicon.ico
    >>>     # assets

    """
    all_static_assets: List[pathlib.Path] = []
    server_path = pathlib.Path(__file__)
    server_static_path = server_path.parent / "static"
    for f in server_static_path.glob("*"):
        all_static_assets.append(f)

    return all_static_assets


def _execute_server_setup_hook(user_app_path: str) -> None:
    """
    Runs the server_setup.py module if present in the application directory.

    """
    server_setup_path = os.path.join(user_app_path, "server_setup.py")
    if os.path.isfile(server_setup_path):
        spec = cast(
            ModuleSpec, importlib.util.spec_from_file_location("server_setup", server_setup_path)
        )
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)  # type: ignore


def _next_localhost_available_port(port_range: Tuple[int, int]) -> int:
    """
    Searches for a free port in a given range on localhost to start the server

    >>> port = _next_localhost_available_port((3005, 3099))

    3005 is the first port to be tested. If it is not available, the port 3006 is tested, and so on.
    """
    for port in range(port_range[0], port_range[1]):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(1)
        result = sock.connect_ex(("127.0.0.1", port))
        sock.close()
        if result != 0:
            return port

    raise OSError(
        f"No free port found to start the server between {port_range[0]} and {port_range[1]} ."
    )
